import os
import torch
import numpy as np
import pandas as pd
from skimage.io import imread
from tqdm import tqdm
from skimage.segmentation import clear_border
from skimage.transform import resize

from biapy.data.post_processing.post_processing import (watershed_by_channels, voronoi_on_mask, 
    measure_morphological_props_and_filter, repare_large_blobs, apply_binary_mask)
from biapy.data.pre_processing import create_instance_channels, create_test_instance_channels, norm_range01
from biapy.utils.util import save_tif, pad_and_reflect
from biapy.utils.matching import matching, wrapper_matching_dataset_lazy
from biapy.engine.metrics import jaccard_index, instance_segmentation_loss, instance_metrics
from biapy.engine.base_workflow import Base_Workflow

class Instance_Segmentation_Workflow(Base_Workflow):
    """
    Instance segmentation workflow where the goal is to assign an unique id, i.e. integer, to each object of the input image.
    More details in `our documentation <https://biapy.readthedocs.io/en/latest/workflows/instance_segmentation.html>`_.  

    Parameters
    ----------
    cfg : YACS configuration
        Running configuration.
    
    Job_identifier : str
        Complete name of the running job.

    device : Torch device
        Device used. 

    args : argpase class
        Arguments used in BiaPy's call. 
    """
    def __init__(self, cfg, job_identifier, device, args, **kwargs):
        super(Instance_Segmentation_Workflow, self).__init__(cfg, job_identifier, device, args, **kwargs)

        self.original_test_path, self.original_test_mask_path = self.prepare_instance_data()

        # Merging the image
        self.all_matching_stats_merge_patches = []
        self.all_matching_stats_merge_patches_post = []
        self.stats['inst_stats_merge_patches'] = None
        self.stats['inst_stats_merge_patches_post'] = None
        # Multi-head: instances + classification
        if self.cfg.MODEL.N_CLASSES > 2:
            self.all_class_stats_merge_patches = []
            self.all_class_stats_merge_patches_post = []
            self.stats['class_stats_merge_patches'] = None
            self.stats['class_stats_merge_patches_post'] = None

        # As 3D stack
        self.all_matching_stats_as_3D_stack = []
        self.all_matching_stats_as_3D_stack_post = []
        self.stats['inst_stats_as_3D_stack'] = None
        self.stats['inst_stats_as_3D_stack_post'] = None
        # Multi-head: instances + classification
        if self.cfg.MODEL.N_CLASSES > 2:
            self.all_class_stats_as_3D_stack = []
            self.all_class_stats_as_3D_stack_post = []
            self.stats['class_stats_as_3D_stack'] = None
            self.stats['class_stats_as_3D_stack_post'] = None

        # Full image
        self.all_matching_stats = []
        self.all_matching_stats_post = []
        self.stats['inst_stats'] = None
        self.stats['inst_stats_post'] = None
        # Multi-head: instances + classification
        if self.cfg.MODEL.N_CLASSES > 2:
            self.all_class_stats = []
            self.all_class_stats_post = []
            self.stats['class_stats'] = None
            self.stats['class_stats_post'] = None
                
        self.instance_ths = {}
        self.instance_ths['TYPE'] = self.cfg.PROBLEM.INSTANCE_SEG.DATA_MW_TH_TYPE
        self.instance_ths['TH_BINARY_MASK'] = self.cfg.PROBLEM.INSTANCE_SEG.DATA_MW_TH_BINARY_MASK
        self.instance_ths['TH_CONTOUR'] = self.cfg.PROBLEM.INSTANCE_SEG.DATA_MW_TH_CONTOUR
        self.instance_ths['TH_FOREGROUND'] = self.cfg.PROBLEM.INSTANCE_SEG.DATA_MW_TH_FOREGROUND
        self.instance_ths['TH_DISTANCE'] = self.cfg.PROBLEM.INSTANCE_SEG.DATA_MW_TH_DISTANCE
        self.instance_ths['TH_POINTS'] = self.cfg.PROBLEM.INSTANCE_SEG.DATA_MW_TH_POINTS 

        # From now on, no modification of the cfg will be allowed
        self.cfg.freeze()

        # Activations for each output channel:
        # channel number : 'activation'
        self.activations = {}
        if self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS == "Dv2":
            self.activations = {'0': 'Linear'}
        elif self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS in ["BC", "BP", "BCM"]:
            self.activations = {':': 'CE_Sigmoid'}
        elif self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS in ["BDv2", "BD"]:
            self.activations = {'0': 'CE_Sigmoid', '1': 'Linear'}
        elif self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS in ["BCD", "BCDv2"]:
            self.activations = {'0': 'CE_Sigmoid', '1': 'CE_Sigmoid', '2': 'Linear'}

        # Multi-head: instances + classification
        if self.cfg.MODEL.N_CLASSES > 2:
            self.activations = [self.activations, {'0': 'Linear'}]
        else:
            self.activations = [self.activations]
            
        # Workflow specific training variables
        self.mask_path = cfg.DATA.TRAIN.GT_PATH
        self.load_Y_val = True

        # Specific instance segmentation post-processing
        if self.cfg.TEST.POST_PROCESSING.VORONOI_ON_MASK or \
            self.cfg.TEST.POST_PROCESSING.MEASURE_PROPERTIES.REMOVE_BY_PROPERTIES.ENABLE or \
            self.cfg.TEST.POST_PROCESSING.REPARE_LARGE_BLOBS_SIZE != -1:
            self.post_processing['instance_post'] = True
        else:
            self.post_processing['instance_post'] = False            
        self.instances_already_created = False 

    def define_metrics(self):
        """
        Definition of self.metrics, self.metric_names and self.loss variables.
        """
        self.metrics = []
        self.metric_names = []
        self.loss = instance_segmentation_loss(
            self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNEL_WEIGHTS,
            self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS, 
            self.cfg.PROBLEM.INSTANCE_SEG.DISTANCE_CHANNEL_MASK,
            self.cfg.MODEL.N_CLASSES
        )

        if self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS == "BC":
            self.metric_names = ["jaccard_index", "jaccard_index"]
        elif self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS == "BCM":
            self.metric_names = ["jaccard_index", "jaccard_index", "jaccard_index"]
        elif self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS == "BP":
            self.metric_names = ["jaccard_index", "jaccard_index"]
        elif self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS == "BD":
            self.metric_names = ["jaccard_index", "L1_distance_channel"]
        elif self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS in ["BCD", 'BCDv2']:
            self.metric_names = ["jaccard_index", "jaccard_index", "L1_distance_channel"]
        elif self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS == "BDv2":
            self.metric_names = ["jaccard_index", "L1_distance_channel"]
        elif self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS == "Dv2":
            self.metric_names = ["L1_distance_channel"]

        # Multi-head: instances + classification    
        if self.cfg.MODEL.N_CLASSES > 2:
            self.metric_names.append("jaccard_index_classes")
            # Used to calculate IoU with the classification results 
            self.jaccard_index_matching = jaccard_index(device="cpu",  
                num_classes=self.cfg.MODEL.N_CLASSES)

        self.metrics = instance_metrics(num_classes=self.cfg.MODEL.N_CLASSES,
            metric_names=self.metric_names, device=self.device, 
            torchvision_models=True if self.cfg.MODEL.SOURCE == "torchvision" else False)

    def metric_calculation(self, output, targets, metric_logger=None):
        """
        Execution of the metrics defined in :func:`~define_metrics` function. 

        Parameters
        ----------
        output : Torch Tensor
            Prediction of the model. 

        targets : Torch Tensor
            Ground truth to compare the prediction with. 

        metric_logger : MetricLogger, optional
            Class to be updated with the new metric(s) value(s) calculated. 
        
        Returns
        -------
        value : float
            Value of the metric for the given prediction. 
        """
        with torch.no_grad():
            out = self.metrics(output, targets)
            first_val = None
            for key, value in out.items():
                value = value.item() if not torch.isnan(value) else 0
                if first_val is None and key == "jaccard_index": 
                    first_val = value
                if metric_logger is not None:
                    metric_logger.meters[key].update(value)
        if first_val is None: first_val = 0
        return first_val

    def instance_seg_process(self, pred, filenames, out_dir, out_dir_post_proc, resolution):
        """
        Instance segmentation workflow engine for test/inference. Process model's prediction to prepare 
        instance segmentation output and calculate metrics. 

        Parameters
        ----------
        pred : 4D/5D Torch tensor
            Model predictions. E.g. ``(num_of_images, y, x, channels)`` for 2D or 
            ``(num_of_images, z, y, x, channels)`` for 3D.

        filenames : List of str
            Predicted image's filenames.

        out_dir : path
            Output directory to save the instances.
            
        out_dir_post_proc : path
            Output directory to save the post-processed instances.
        """
        #############################
        ### INSTANCE SEGMENTATION ###
        #############################
        if not self.instances_already_created: 

            # Multi-head: instances + classification
            if self.cfg.MODEL.N_CLASSES > 2:
                class_channel = np.expand_dims(pred[...,-1], -1)
                pred = pred[...,:-1] 

            print("Creating instances with watershed . . .")
            w_dir = os.path.join(self.cfg.PATHS.WATERSHED_DIR, filenames[0])
            check_wa = w_dir if self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHECK_MW else None
            
            w_pred = watershed_by_channels(pred, self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS, ths=self.instance_ths, 
                remove_before=self.cfg.PROBLEM.INSTANCE_SEG.DATA_REMOVE_BEFORE_MW, thres_small_before=self.cfg.PROBLEM.INSTANCE_SEG.DATA_REMOVE_SMALL_OBJ_BEFORE,
                seed_morph_sequence=self.cfg.PROBLEM.INSTANCE_SEG.SEED_MORPH_SEQUENCE, seed_morph_radius=self.cfg.PROBLEM.INSTANCE_SEG.SEED_MORPH_RADIUS, 
                erode_and_dilate_foreground=self.cfg.PROBLEM.INSTANCE_SEG.ERODE_AND_DILATE_FOREGROUND, fore_erosion_radius=self.cfg.PROBLEM.INSTANCE_SEG.FORE_EROSION_RADIUS, 
                fore_dilation_radius=self.cfg.PROBLEM.INSTANCE_SEG.FORE_DILATION_RADIUS, rmv_close_points=self.cfg.TEST.POST_PROCESSING.REMOVE_CLOSE_POINTS, 
                remove_close_points_radius=self.cfg.TEST.POST_PROCESSING.REMOVE_CLOSE_POINTS_RADIUS[0], resolution=resolution, save_dir=check_wa,
                watershed_by_2d_slices=self.cfg.PROBLEM.INSTANCE_SEG.WATERSHED_BY_2D_SLICES)

            # Multi-head: instances + classification
            if self.cfg.MODEL.N_CLASSES > 2:
                print("Adapting class channel . . .")
                labels = np.unique(w_pred)[1:]
                new_class_channel = np.zeros(w_pred.shape, dtype=w_pred.dtype)
                # Classify each instance counting the most prominent class of all the pixels that compose it
                for l in labels:
                    instance_classes, instance_classes_count = np.unique(class_channel[w_pred == l], return_counts=True)
                    
                    # Remove background
                    if instance_classes[0] == 0:
                        instance_classes = instance_classes[1:]
                        instance_classes_count = instance_classes_count[1:]

                    if len(instance_classes) > 0:
                        label_selected = int(instance_classes[np.argmax(instance_classes_count)])
                    else: # Label by default with class 1 in case there was no class info
                        label_selected = 1 
                    new_class_channel = np.where(w_pred == l, label_selected, new_class_channel) 

                class_channel = new_class_channel
                class_channel = class_channel.squeeze()
                del new_class_channel
                save_tif(np.expand_dims(np.concatenate([np.expand_dims(w_pred.squeeze(),-1), np.expand_dims(class_channel,-1)],axis=-1),0), 
                    out_dir, filenames, verbose=self.cfg.TEST.VERBOSE)
            else:
                save_tif(np.expand_dims(np.expand_dims(w_pred,-1),0), out_dir, filenames, verbose=self.cfg.TEST.VERBOSE)

            # Add extra dimension if working in 2D
            if w_pred.ndim == 2:
                w_pred = np.expand_dims(w_pred,0) 
        else:
            w_pred = pred.squeeze()
            if w_pred.ndim == 2: w_pred = np.expand_dims(w_pred,0)

        results = None
        results_class = None
        if self.cfg.TEST.MATCHING_STATS and (self.cfg.DATA.TEST.LOAD_GT or self.cfg.DATA.TEST.USE_VAL_AS_TEST):
            print("Calculating matching stats . . .")

            # Need to load instance labels, as Y are binary channels used for IoU calculation
            if self.cfg.TEST.ANALIZE_2D_IMGS_AS_3D_STACK and len(self.test_filenames) == w_pred.shape[0]:
                del self._Y
                _Y = np.zeros(w_pred.shape, dtype=w_pred.dtype)
                for i in range(len(self.test_filenames)):
                    test_file = os.path.join(self.original_test_mask_path, self.test_filenames[i])
                    _Y[i] = imread(test_file).squeeze()
            else:
                test_file = os.path.join(self.original_test_mask_path, self.test_filenames[self.f_numbers[0]])
                _Y = imread(test_file).squeeze()

            # Multi-head: instances + classification
            if self.cfg.MODEL.N_CLASSES > 2:
                # Channel check
                error_shape = None
                if self.cfg.PROBLEM.NDIM == "2D" and _Y.ndim != 3:
                    error_shape = (256,256,2)
                elif self.cfg.PROBLEM.NDIM == "3D" and _Y.ndim != 4:
                    error_shape = (40,256,256,2)
                if error_shape is not None:
                    raise ValueError(f"Image {test_file} wrong dimension. In instance segmentation, when 'MODEL.N_CLASSES' are "
                        f"more than 2 labels need to have two channels, e.g. {error_shape}, containing the instance "
                        "segmentation map (first channel) and classification map (second channel).")
                
                # Ensure channel position
                if self.cfg.PROBLEM.NDIM == "2D":
                    if _Y.shape[0] <= 3: _Y = _Y.transpose((1,2,0)) 
                else:    
                    min_val = min(_Y.shape)
                    channel_pos = _Y.shape.index(min_val)
                    if channel_pos != 3 and _Y.shape[channel_pos] <= 4:
                        new_pos = [x for x in range(4) if x != channel_pos]+[channel_pos,]
                        _Y = _Y.transpose(new_pos)

                # Separate instance and classification channels
                _Y_classes = _Y[...,1] # Classes 
                _Y = _Y[...,0] # Instances 

                # Measure class IoU
                class_iou = self.jaccard_index_matching(torch.as_tensor(class_channel.squeeze().astype(np.int32)),
                    torch.as_tensor(_Y_classes.squeeze().astype(np.int32)))
                class_iou = class_iou.item() if not torch.isnan(class_iou) else 0
                print(f"Class IoU: {class_iou}")
                results_class = class_iou

            if _Y.ndim == 2: 
                _Y = np.expand_dims(_Y,0)

            # For torchvision models that resize need to rezise the images 
            if w_pred.shape != _Y.shape:
                _Y = resize(_Y, w_pred.shape, order=0)

            # Convert instances to integer
            if _Y.dtype == np.float32: _Y = _Y.astype(np.uint32)
            if _Y.dtype == np.float64: _Y = _Y.astype(np.uint64)

            diff_ths_colored_img = abs(len(self.cfg.TEST.MATCHING_STATS_THS_COLORED_IMG) - len(self.cfg.TEST.MATCHING_STATS_THS))
            colored_img_ths = self.cfg.TEST.MATCHING_STATS_THS_COLORED_IMG+[-1]*diff_ths_colored_img

            results = matching(_Y, w_pred, thresh=self.cfg.TEST.MATCHING_STATS_THS, report_matches=True)
            for i in range(len(results)):
                # Extract TPs, FPs and FNs from the resulting matching data structure 
                r_stats = results[i] 
                thr = r_stats['thresh']

                # TP and FN
                gt_ids = r_stats['gt_ids'][1:]
                matched_pairs = r_stats['matched_pairs']
                gt_match = [x[0] for x in matched_pairs]
                gt_unmatch = [x for x in gt_ids if x not in gt_match]
                matched_scores = list(r_stats['matched_scores'])+[0 for _ in gt_unmatch]
                pred_match = [x[1] for x in matched_pairs]+[-1 for _ in gt_unmatch]
                tag = ["TP" if score >= thr else "FN" for score in matched_scores]

                # FPs
                pred_ids = r_stats['pred_ids'][1:]
                fp_instances = [x for x in pred_ids if x not in pred_match]
                fp_instances += [pred_id for score, pred_id in zip(matched_scores, pred_match) if score < thr]

                # Save csv files
                df = pd.DataFrame(zip(gt_match+gt_unmatch, pred_match, matched_scores, tag), columns =['gt_id', 'pred_id', 'iou', 'tag'])
                df = df.sort_values(by=['gt_id'])  
                df_fp = pd.DataFrame(zip(fp_instances), columns =['pred_id'])

                os.makedirs(self.cfg.PATHS.RESULT_DIR.INST_ASSOC_POINTS, exist_ok=True)
                df.to_csv(os.path.join(self.cfg.PATHS.RESULT_DIR.INST_ASSOC_POINTS, os.path.splitext(filenames[0])[0]+'_th_{}_gt_assoc.csv'.format(thr)), index=False)
                df_fp.to_csv(os.path.join(self.cfg.PATHS.RESULT_DIR.INST_ASSOC_POINTS, os.path.splitext(filenames[0])[0]+'_th_{}_fp.csv'.format(thr)), index=False)
                del r_stats['matched_scores']; del r_stats['matched_tps']; del r_stats['matched_pairs']; del r_stats['pred_ids']; del r_stats['gt_ids']
                print("DatasetMatching: {}".format(r_stats))

                if colored_img_ths[i] != -1 and colored_img_ths[i] == thr:
                    print("Creating the image with a summary of detected points and false positives with colors . . .")
                    colored_result = np.zeros(w_pred.shape+(3,), dtype=np.uint8)

                    print("Painting TPs and FNs . . .")
                    for j in tqdm(range(len(gt_match))):
                        color = (0,255,0) if tag[j] == "TP" else (255,0,0) # Green or red
                        colored_result[np.where(_Y == gt_match[j])] = color
                    for j in tqdm(range(len(gt_unmatch))):
                        colored_result[np.where(_Y == gt_unmatch[j])] = (255,0,0) # Red

                    print("Painting FPs . . .")
                    for j in tqdm(range(len(fp_instances))):
                        colored_result[np.where(w_pred == fp_instances[j])] = (0,0,255) # Blue

                    save_tif(np.expand_dims(colored_result,0), self.cfg.PATHS.RESULT_DIR.INST_ASSOC_POINTS,
                            [os.path.splitext(filenames[0])[0]+'_th_{}.tif'.format(thr)], verbose=self.cfg.TEST.VERBOSE)          
                    del colored_result

        ###################
        # Post-processing #
        ###################
        if self.cfg.TEST.POST_PROCESSING.REPARE_LARGE_BLOBS_SIZE != -1:
            w_pred = repare_large_blobs(w_pred, self.cfg.TEST.POST_PROCESSING.REPARE_LARGE_BLOBS_SIZE)

        if self.cfg.TEST.POST_PROCESSING.MEASURE_PROPERTIES.ENABLE or \
            self.cfg.TEST.POST_PROCESSING.MEASURE_PROPERTIES.REMOVE_BY_PROPERTIES.ENABLE:
            w_pred, d_result = measure_morphological_props_and_filter(w_pred.squeeze(), resolution, 
                filter_instances=self.cfg.TEST.POST_PROCESSING.MEASURE_PROPERTIES.REMOVE_BY_PROPERTIES.ENABLE,
                properties=self.cfg.TEST.POST_PROCESSING.MEASURE_PROPERTIES.REMOVE_BY_PROPERTIES.PROPS, 
                prop_values=self.cfg.TEST.POST_PROCESSING.MEASURE_PROPERTIES.REMOVE_BY_PROPERTIES.VALUES,
                comp_signs=self.cfg.TEST.POST_PROCESSING.MEASURE_PROPERTIES.REMOVE_BY_PROPERTIES.SIGN)

            # Save all instance stats            
            if self.cfg.PROBLEM.NDIM == "2D":
                df = pd.DataFrame(zip(np.array(d_result['labels'], dtype=np.uint64), list(d_result['centers'][:,0]), list(d_result['centers'][:,1]), 
                    d_result['npixels'], d_result['areas'], d_result['circularities'], d_result['diameters'], d_result['perimeters'], d_result['elongations'],
                    d_result['comment'], d_result['conditions']), columns=['label', 'axis-0', 'axis-1', 'npixels', 'area', 'circularity', 'diameter', 
                    'perimeter', 'elongation', 'comment', 'conditions'])
            else:
                df = pd.DataFrame(zip(np.array(d_result['labels'], dtype=np.uint64), list(d_result['centers'][:,0]), list(d_result['centers'][:,1]), 
                    list(d_result['centers'][:,2]), d_result['npixels'], d_result['areas'], d_result['circularities'], d_result['diameters'], 
                    d_result['perimeters'], d_result['comment'], d_result['conditions']), columns=['label', 'axis-0', 'axis-1', 
                    'axis-2', 'npixels', 'volume', 'sphericity', 'diameter', 'perimeter (surface area)', 'comment', 'conditions'])
            df = df.sort_values(by=['label'])   
            df.to_csv(os.path.join(out_dir, os.path.splitext(filenames[0])[0]+'_full_stats.csv'), index=False)
            # Save only remain instances stats
            df = df[df["comment"].str.contains("Strange")==False] 
            os.makedirs(out_dir_post_proc, exist_ok=True)
            df.to_csv(os.path.join(out_dir_post_proc, os.path.splitext(filenames[0])[0]+'_filtered_stats.csv'), index=False)
            del df

        if self.cfg.TEST.POST_PROCESSING.VORONOI_ON_MASK:
            w_pred = voronoi_on_mask(w_pred, pred, th=self.cfg.TEST.POST_PROCESSING.VORONOI_TH, verbose=self.cfg.TEST.VERBOSE)
        del pred

        if self.cfg.TEST.POST_PROCESSING.CLEAR_BORDER:
            print("Clearing borders . . .")
            w_pred = clear_border(w_pred)

        results_post_proc = None
        results_class_post_proc = None
        if self.post_processing['instance_post']:
            # Multi-head: instances + classification
            if self.cfg.MODEL.N_CLASSES > 2:
                w_pred = w_pred.squeeze()
                class_channel = np.where(w_pred>0, class_channel, 0) # Adapt changes to post-processed w_pred
                save_tif(np.expand_dims(np.concatenate([np.expand_dims(w_pred,-1), np.expand_dims(class_channel,-1)],axis=-1),0), 
                    out_dir_post_proc, filenames, verbose=self.cfg.TEST.VERBOSE)
            else:
                save_tif(np.expand_dims(np.expand_dims(w_pred,-1),0), out_dir_post_proc, filenames, verbose=self.cfg.TEST.VERBOSE)

            if self.cfg.TEST.MATCHING_STATS and (self.cfg.DATA.TEST.LOAD_GT or self.cfg.DATA.TEST.USE_VAL_AS_TEST):
                # Multi-head: instances + classification
                if self.cfg.MODEL.N_CLASSES > 2:
                    # Measure class IoU
                    class_iou = self.jaccard_index_matching(torch.as_tensor(class_channel.squeeze().astype(np.int32)),
                        torch.as_tensor(_Y_classes.squeeze().astype(np.int32)))
                    class_iou = class_iou.item() if not torch.isnan(class_iou) else 0
                    print(f"Class IoU (post-processing): {class_iou}")
                    results_class_post_proc = class_iou
                    
                # Add extra dimension if working in 2D
                if w_pred.ndim == 2:
                    w_pred = np.expand_dims(w_pred,0)

                print("Calculating matching stats after post-processing . . .")
                results_post_proc = matching(_Y, w_pred, thresh=self.cfg.TEST.MATCHING_STATS_THS, report_matches=True)
                
                for i in range(len(results_post_proc)):
                    # Extract TPs, FPs and FNs from the resulting matching data structure 
                    r_stats = results_post_proc[i] 
                    thr = r_stats['thresh']

                    # TP and FN
                    gt_ids = r_stats['gt_ids'][1:]
                    matched_pairs = r_stats['matched_pairs']
                    gt_match = [x[0] for x in matched_pairs]
                    gt_unmatch = [x for x in gt_ids if x not in gt_match]
                    matched_scores = list(r_stats['matched_scores'])+[0 for _ in gt_unmatch]
                    pred_match = [x[1] for x in matched_pairs]+[-1 for _ in gt_unmatch]
                    tag = ["TP" if score >= thr else "FN" for score in matched_scores]

                    # FPs
                    pred_ids = r_stats['pred_ids'][1:]
                    fp_instances = [x for x in pred_ids if x not in pred_match]
                    fp_instances += [pred_id for score, pred_id in zip(matched_scores, pred_match) if score < thr]

                    # Save csv files
                    df = pd.DataFrame(zip(gt_match+gt_unmatch, pred_match, matched_scores, tag), columns =['gt_id', 'pred_id', 'iou', 'tag'])
                    df = df.sort_values(by=['gt_id'])  
                    df_fp = pd.DataFrame(zip(fp_instances), columns =['pred_id'])

                    os.makedirs(self.cfg.PATHS.RESULT_DIR.INST_ASSOC_POINTS, exist_ok=True)
                    df.to_csv(os.path.join(self.cfg.PATHS.RESULT_DIR.INST_ASSOC_POINTS, os.path.splitext(filenames[0])[0]+'_post-proc_th_{}_gt_assoc.csv'.format(thr)), index=False)
                    df_fp.to_csv(os.path.join(self.cfg.PATHS.RESULT_DIR.INST_ASSOC_POINTS, os.path.splitext(filenames[0])[0]+'_post-proc_th_{}_fp.csv'.format(thr)), index=False)
                    del r_stats['matched_scores']; del r_stats['matched_tps']; del r_stats['matched_pairs']; del r_stats['pred_ids']; del r_stats['gt_ids']
                    print("DatasetMatching: {}".format(r_stats))

                    if colored_img_ths[i] != -1 and colored_img_ths[i] == thr:
                        print("Creating the image with a summary of detected points and false positives with colors . . .")
                        colored_result = np.zeros(w_pred.shape+(3,), dtype=np.uint8)

                        print("Painting TPs and FNs . . .")
                        for j in tqdm(range(len(gt_match))):
                            color = (0,255,0) if tag[j] == "TP" else (255,0,0) # Green or red
                            colored_result[np.where(_Y == gt_match[j])] = color
                        for j in tqdm(range(len(gt_unmatch))):
                            colored_result[np.where(_Y == gt_unmatch[j])] = (255,0,0) # Red
                            
                        print("Painting FPs . . .")
                        for j in tqdm(range(len(fp_instances))):
                            colored_result[np.where(w_pred == fp_instances[j])] = (0,0,255) # Blue

                        save_tif(np.expand_dims(colored_result,0), self.cfg.PATHS.RESULT_DIR.INST_ASSOC_POINTS,
                                [os.path.splitext(filenames[0])[0]+'_post-proc_th_{}.tif'.format(thr)], verbose=self.cfg.TEST.VERBOSE)          
                        del colored_result

        return results, results_post_proc, results_class, results_class_post_proc

    def process_sample(self, norm):
        """
        Function to process a sample in the inference phase. 

        Parameters
        ----------
        norm : List of dicts
            Normalization used during training. Required to denormalize the predictions of the model.
        """
        if self.cfg.MODEL.SOURCE != "torchvision":
            self.instances_already_created = False
            super().process_sample(norm)
        else:
            self.instances_already_created = True
            # Data channel check
            if self.cfg.DATA.PATCH_SIZE[-1] != self._X.shape[-1]:
                raise ValueError("Channel of the DATA.PATCH_SIZE given {} does not correspond with the loaded image {}. "
                    "Please, check the channels of the images!".format(self.cfg.DATA.PATCH_SIZE[-1], self._X.shape[-1]))

            ##################
            ### FULL IMAGE ###
            ##################
            if self.cfg.TEST.FULL_IMG:
                # Make the prediction
                with torch.cuda.amp.autocast():
                    pred = self.model_call_func(self._X)
                del self._X 

                if self.cfg.TEST.POST_PROCESSING.APPLY_MASK:
                    pred = apply_binary_mask(pred, self.cfg.DATA.TEST.BINARY_MASKS)
    
                self.after_full_image(pred)

    def after_merge_patches(self, pred):
        """
        Steps need to be done after merging all predicted patches into the original image.

        Parameters
        ----------
        pred : Torch Tensor
            Model prediction.
        """
        if not self.cfg.TEST.ANALIZE_2D_IMGS_AS_3D_STACK:
            resolution = self.cfg.DATA.TEST.RESOLUTION if len(self.cfg.DATA.TEST.RESOLUTION) == 2 else self.cfg.DATA.TEST.RESOLUTION[1:]
            r, r_post, rcls, rcls_post = self.instance_seg_process(pred, self.processing_filenames, self.cfg.PATHS.RESULT_DIR.PER_IMAGE_INSTANCES,
                self.cfg.PATHS.RESULT_DIR.PER_IMAGE_POST_PROCESSING, resolution)        
            if r is not None:
                self.all_matching_stats_merge_patches.append(r)
            if r_post is not None:
                self.all_matching_stats_merge_patches_post.append(r_post)
            if rcls is not None:
                self.all_class_stats_merge_patches.append(rcls)
            if rcls_post is not None:
                self.all_class_stats_merge_patches_post.append(rcls_post)    

    def after_merge_patches_by_chunks_proccess_patch(self, filename):
        """
        Place any code that needs to be done after merging all predicted patches into the original image
        but in the process made chunk by chunk. This function will operate patch by patch defined by 
        ``DATA.PATCH_SIZE``.

        Parameters
        ----------
        filename : List of str
            Filename of the predicted image H5/Zarr.  
        """
        pass

    def after_full_image(self, pred):
        """
        Steps that must be executed after generating the prediction by supplying the entire image to the model.

        Parameters
        ----------
        pred : Torch Tensor
            Model prediction.
        """
        if not self.cfg.TEST.ANALIZE_2D_IMGS_AS_3D_STACK:
            resolution = self.cfg.DATA.TEST.RESOLUTION if len(self.cfg.DATA.TEST.RESOLUTION) == 2 else self.cfg.DATA.TEST.RESOLUTION[1:]
            r, r_post, rcls, rcls_post = self.instance_seg_process(pred, self.processing_filenames, self.cfg.PATHS.RESULT_DIR.FULL_IMAGE_INSTANCES,
                self.cfg.PATHS.RESULT_DIR.FULL_IMAGE_POST_PROCESSING, resolution)  
            if r is not None:
                self.all_matching_stats.append(r)
            if r_post is not None:
                self.all_matching_stats_post.append(r_post)
            if rcls is not None:
                self.all_class_stats.append(rcls)
            if rcls_post is not None:
                self.all_class_stats_post.append(rcls_post) 

    def after_all_images(self):
        """
        Steps that must be done after predicting all images. 
        """
        super().after_all_images()
        if self.cfg.TEST.ANALIZE_2D_IMGS_AS_3D_STACK:
            print("Analysing all images as a 3D stack . . .")    
            if type(self.all_pred) is list:
                self.all_pred = np.concatenate(self.all_pred)
            resolution = self.cfg.DATA.TEST.RESOLUTION if len(self.cfg.DATA.TEST.RESOLUTION) == 3 else (self.cfg.DATA.TEST.RESOLUTION[0],)+self.cfg.DATA.TEST.RESOLUTION
            r, r_post, rcls, rcls_post = self.instance_seg_process(self.all_pred, ["3D_stack.tif"], self.cfg.PATHS.RESULT_DIR.AS_3D_STACK,
                self.cfg.PATHS.RESULT_DIR.AS_3D_STACK_POST_PROCESSING, resolution)
            if r is not None:
                self.all_matching_stats_as_3D_stack.append(r)
            if r_post is not None:
                self.all_matching_stats_as_3D_stack_post.append(r_post)
            if rcls is not None:
                self.all_class_stats_as_3D_stack.append(rcls)
            if rcls_post is not None:
                self.all_class_stats_as_3D_stack_post.append(rcls_post) 

    def normalize_stats(self, image_counter): 
        """
        Normalize statistics.  

        Parameters
        ----------
        image_counter : int
            Number of images to average the metrics.
        """
        super().normalize_stats(image_counter) 

        if (self.cfg.DATA.TEST.LOAD_GT or self.cfg.DATA.TEST.USE_VAL_AS_TEST):
            if self.cfg.TEST.MATCHING_STATS:
                # Merge patches
                if len(self.all_matching_stats_merge_patches) > 0:
                    self.stats['inst_stats_merge_patches'] = wrapper_matching_dataset_lazy(self.all_matching_stats_merge_patches, self.cfg.TEST.MATCHING_STATS_THS)
                # As 3D stack
                if len(self.all_matching_stats_as_3D_stack) > 0:
                    self.stats['inst_stats_as_3D_stack'] = wrapper_matching_dataset_lazy(self.all_matching_stats_as_3D_stack, self.cfg.TEST.MATCHING_STATS_THS)
                # Full image
                if len(self.all_matching_stats) > 0:
                    self.stats['inst_stats'] = wrapper_matching_dataset_lazy(self.all_matching_stats, self.cfg.TEST.MATCHING_STATS_THS)
                if self.post_processing['instance_post']:
                    # Merge patches
                    if len(self.all_matching_stats_merge_patches_post) > 0:
                        self.stats['inst_stats_merge_patches_post'] = wrapper_matching_dataset_lazy(self.all_matching_stats_merge_patches_post, self.cfg.TEST.MATCHING_STATS_THS)
                    # As 3D stack
                    if len(self.all_matching_stats_as_3D_stack_post) > 0:
                        self.stats['inst_stats_as_3D_stack_post'] = wrapper_matching_dataset_lazy(self.all_matching_stats_as_3D_stack_post, self.cfg.TEST.MATCHING_STATS_THS)
                    # Full image
                    if len(self.all_matching_stats_post) > 0:
                        self.stats['inst_stats_post'] = wrapper_matching_dataset_lazy(self.all_matching_stats_post, self.cfg.TEST.MATCHING_STATS_THS)
                
                # Multi-head: instances + classification
                if self.cfg.MODEL.N_CLASSES > 2:
                    # Merge patches
                    if len(self.all_class_stats_merge_patches) > 0:
                        self.stats['class_stats_merge_patches'] = np.mean(self.all_class_stats_merge_patches)
                    # As 3D stack    
                    if len(self.all_class_stats_as_3D_stack) > 0:
                        self.stats['class_stats_as_3D_stack'] = np.mean(self.all_class_stats_as_3D_stack)
                    # Full image
                    if len(self.all_class_stats) > 0:
                        self.stats['class_stats'] = np.mean(self.all_class_stats)

                    if self.post_processing['instance_post']:
                        # Merge patches
                        if len(self.all_class_stats_merge_patches_post) > 0:
                            self.stats['class_stats_merge_patches_post'] = np.mean(self.all_class_stats_merge_patches_post) 
                        # As 3D stack   
                        if len(self.all_class_stats_as_3D_stack_post) > 0:
                            self.stats['class_stats_as_3D_stack_post'] = np.mean(self.all_class_stats_as_3D_stack_post)
                        # Full image 
                        if len(self.all_class_stats_post) > 0:
                            self.stats['class_stats_post'] = np.mean(self.all_class_stats_post)

    def print_stats(self, image_counter):
        """
        Print statistics.  

        Parameters
        ----------
        image_counter : int
            Number of images to call ``normalize_stats``.
        """
        if self.cfg.MODEL.SOURCE != "torchvision":
            super().print_stats(image_counter)
            super().print_post_processing_stats()

            print("Instance segmentation specific metrics:")
            if self.cfg.TEST.MATCHING_STATS and (self.cfg.DATA.TEST.LOAD_GT or self.cfg.DATA.TEST.USE_VAL_AS_TEST) :
                for i in range(len(self.cfg.TEST.MATCHING_STATS_THS)):
                    print("IoU TH={}".format(self.cfg.TEST.MATCHING_STATS_THS[i]))
                    # Merge patches 
                    if self.stats['inst_stats_merge_patches'] is not None:
                        print("      Merge patches:")
                        print(f"      {self.stats['inst_stats_merge_patches'][i]}")
                    # As 3D stack
                    if self.stats['inst_stats_as_3D_stack'] is not None:
                        print("      As 3D stack:")
                        print(f"      {self.stats['inst_stats_as_3D_stack'][i]}")
                    # Full image
                    if self.stats['inst_stats'] is not None:
                        print("      Full image:")
                        print(f"      {self.stats['inst_stats'][i]}")
                    if self.post_processing['instance_post']:
                        print("IoU (post-processing) TH={}".format(self.cfg.TEST.MATCHING_STATS_THS[i]))
                        # Merge patches 
                        if self.stats['inst_stats_merge_patches_post'] is not None:
                            print("      Merge patches (post-processing):")
                            print(f"      {self.stats['inst_stats_merge_patches_post'][i]}")
                        # As 3D stack
                        if self.stats['inst_stats_as_3D_stack_post'] is not None:
                            print("      As 3D stack (post-processing):")
                            print(f"      {self.stats['inst_stats_as_3D_stack_post'][i]}")
                        # Full image
                        if self.stats['inst_stats_post'] is not None:
                            print("      Full image (post-processing):")
                            print(f"      {self.stats['inst_stats_post'][i]}")

                # Multi-head: instances + classification
                if self.cfg.MODEL.N_CLASSES > 2:
                    # Merge patches
                    if self.stats['class_stats_merge_patches'] is not None:
                        print(f"      Merge patches classification IoU: {self.stats['class_stats_merge_patches']}")
                    # As 3D stack    
                    if self.stats['class_stats_as_3D_stack'] is not None:
                        print(f"      As 3D stack classification IoU: {self.stats['class_stats_as_3D_stack']}")
                    # Full image
                    if self.stats['class_stats'] is not None:
                        print(f"      Full image classification IoU: {self.stats['class_stats']}")

                    if self.post_processing['instance_post']:
                        # Merge patches
                        if self.stats['class_stats_merge_patches_post'] is not None:
                            print(f"      Merge patches classification IoU (post-processing): {self.stats['class_stats_merge_patches_post']}")
                        # As 3D stack    
                        if self.stats['class_stats_as_3D_stack_post'] is not None:
                            print(f"      As 3D stack classification IoU (post-processing): {self.stats['class_stats_as_3D_stack_post']}")
                        # Full image
                        if self.stats['class_stats_post'] is not None:
                            print(f"      Full image classification IoU (post-processing): {self.stats['class_stats_post']}")

    def prepare_instance_data(self):
        """
        Creates instance segmentation ground truth images to train the model based on the ground truth instances provided.
        They will be saved in a separate folder in the root path of the ground truth. 
        """
        print("###########################")
        print("#  PREPARE INSTANCE DATA  #")
        print("###########################")
        original_test_path, original_test_mask_path = None, None

        # Create selected channels for train data
        if (self.cfg.TRAIN.ENABLE or self.cfg.DATA.TEST.USE_VAL_AS_TEST) and (not os.path.isdir(self.cfg.DATA.TRAIN.INSTANCE_CHANNELS_DIR) or \
            not os.path.isdir(self.cfg.DATA.TRAIN.INSTANCE_CHANNELS_MASK_DIR)):
            print("You select to create {} channels from given instance labels and no file is detected in {}. "
                    "So let's prepare the data. Notice that, if you do not modify 'DATA.TRAIN.INSTANCE_CHANNELS_DIR' "
                    "path, this process will be done just once!".format(self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS,
                    self.cfg.DATA.TRAIN.INSTANCE_CHANNELS_DIR))
            create_instance_channels(self.cfg)

        # Create selected channels for val data
        if self.cfg.TRAIN.ENABLE and not self.cfg.DATA.VAL.FROM_TRAIN and (not os.path.isdir(self.cfg.DATA.VAL.INSTANCE_CHANNELS_DIR) or \
            not os.path.isdir(self.cfg.DATA.VAL.INSTANCE_CHANNELS_MASK_DIR)):
            print("You select to create {} channels from given instance labels and no file is detected in {}. "
                    "So let's prepare the data. Notice that, if you do not modify 'DATA.VAL.INSTANCE_CHANNELS_DIR' "
                    "path, this process will be done just once!".format(self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS,
                    self.cfg.DATA.VAL.INSTANCE_CHANNELS_DIR))
            create_instance_channels(self.cfg, data_type='val')

        # Create selected channels for test data once
        if self.cfg.TEST.ENABLE and not self.cfg.DATA.TEST.USE_VAL_AS_TEST and (not os.path.isdir(self.cfg.DATA.TEST.INSTANCE_CHANNELS_DIR) or \
            (not os.path.isdir(self.cfg.DATA.TEST.INSTANCE_CHANNELS_MASK_DIR) and self.cfg.DATA.TEST.LOAD_GT)) and not self.cfg.TEST.BY_CHUNKS.ENABLE:
            print("You select to create {} channels from given instance labels and no file is detected in {}. "
                    "So let's prepare the data. Notice that, if you do not modify 'DATA.TEST.INSTANCE_CHANNELS_DIR' "
                    "path, this process will be done just once!".format(self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS,
                    self.cfg.DATA.TEST.INSTANCE_CHANNELS_MASK_DIR))
            create_test_instance_channels(self.cfg)

        opts = []
        if self.cfg.TRAIN.ENABLE:
            print("DATA.TRAIN.PATH changed from {} to {}".format(self.cfg.DATA.TRAIN.PATH, self.cfg.DATA.TRAIN.INSTANCE_CHANNELS_DIR))
            print("DATA.TRAIN.GT_PATH changed from {} to {}".format(self.cfg.DATA.TRAIN.GT_PATH, self.cfg.DATA.TRAIN.INSTANCE_CHANNELS_MASK_DIR))
            opts.extend(['DATA.TRAIN.PATH', self.cfg.DATA.TRAIN.INSTANCE_CHANNELS_DIR,
                        'DATA.TRAIN.GT_PATH', self.cfg.DATA.TRAIN.INSTANCE_CHANNELS_MASK_DIR])
        if not self.cfg.DATA.VAL.FROM_TRAIN:
            print("DATA.VAL.PATH changed from {} to {}".format(self.cfg.DATA.VAL.PATH, self.cfg.DATA.VAL.INSTANCE_CHANNELS_DIR))
            print("DATA.VAL.GT_PATH changed from {} to {}".format(self.cfg.DATA.VAL.GT_PATH, self.cfg.DATA.VAL.INSTANCE_CHANNELS_MASK_DIR))
            opts.extend(['DATA.VAL.PATH', self.cfg.DATA.VAL.INSTANCE_CHANNELS_DIR,
                        'DATA.VAL.GT_PATH', self.cfg.DATA.VAL.INSTANCE_CHANNELS_MASK_DIR])
        if self.cfg.TEST.ENABLE and not self.cfg.DATA.TEST.USE_VAL_AS_TEST and not self.cfg.TEST.BY_CHUNKS.ENABLE:
            print("DATA.TEST.PATH changed from {} to {}".format(self.cfg.DATA.TEST.PATH, self.cfg.DATA.TEST.INSTANCE_CHANNELS_DIR))
            opts.extend(['DATA.TEST.PATH', self.cfg.DATA.TEST.INSTANCE_CHANNELS_DIR])
            if self.cfg.DATA.TEST.LOAD_GT:
                print("DATA.TEST.GT_PATH changed from {} to {}".format(self.cfg.DATA.TEST.GT_PATH, self.cfg.DATA.TEST.INSTANCE_CHANNELS_MASK_DIR))
                opts.extend(['DATA.TEST.GT_PATH', self.cfg.DATA.TEST.INSTANCE_CHANNELS_MASK_DIR])
        original_test_path = self.cfg.DATA.TEST.PATH
        original_test_mask_path = self.cfg.DATA.TEST.GT_PATH
        self.cfg.merge_from_list(opts)

        return original_test_path, original_test_mask_path

    def torchvision_model_call(self, in_img, is_train=False):
        """
        Call a regular Pytorch model.

        Parameters
        ----------
        in_img : Tensor
            Input image to pass through the model.

        is_train : bool, optional
            Whether if the call is during training or inference. 

        Returns
        -------
        prediction : Tensor 
            Image prediction. 
        """
        filename, file_extension = os.path.splitext(self.processing_filenames[0])

        # Convert first to 0-255 range if uint16
        if in_img.dtype == torch.float32:
            if torch.max(in_img) > 1:
                in_img = (norm_range01(in_img, torch.uint8)[0]*255).to(torch.uint8)
            in_img = in_img.to(torch.uint8)
        
        # Apply TorchVision pre-processing
        in_img = self.torchvision_preprocessing(in_img)
        pred = self.model(in_img)
        masks = pred[0]['masks'].cpu().numpy().transpose(0,2,3,1)
        if masks.shape[0] != 0:
            masks = np.argmax(pred[0]['masks'].cpu().numpy().transpose(0,2,3,1), axis=0)
        else:
            masks = torch.ones((1,)+pred[0]['masks'].cpu().numpy().transpose(0,2,3,1).shape[1:], dtype=torch.uint8)

        if not is_train and masks.shape[0] != 0:
            # Extract each output from MaskRCNN
            bboxes = pred[0]['boxes'].cpu().numpy().astype(np.uint16)
            labels = pred[0]['labels'].cpu().numpy()
            scores = pred[0]['scores'].cpu().numpy()
            
            # Save all info in a csv file
            df = pd.DataFrame(zip(labels, scores, bboxes[:,0],bboxes[:,1],bboxes[:,2],bboxes[:,3]), 
                columns = ['label', 'scores', 'x1', 'y1', 'x2', 'y2'])
            df = df.sort_values(by=['label']) 
            df.to_csv(os.path.join(self.cfg.PATHS.RESULT_DIR.FULL_IMAGE, filename+".csv"), index=False)

            # Save masks
            save_tif(np.expand_dims(masks,0), self.cfg.PATHS.RESULT_DIR.FULL_IMAGE, self.processing_filenames, 
                verbose=self.cfg.TEST.VERBOSE)    

        # Actually to allow training this should return a Torch Tensor and not a Numpy array. For instance, 
        # segmentation training is disabled, due to the absence of generators that contain bboxes, so this can be left 
        # returning a Numpy array. This will only be called in process_sample inference function and for full image setting
        return np.expand_dims(masks.squeeze(),0) 


import math
import torch
import torch.nn as nn
from timm.models.vision_transformer import Block
from typing import List

from biapy.models.blocks import DoubleConvBlock, ConvBlock
from biapy.models.tr_layers import PatchEmbed

class UNETR(nn.Module):
    """
    UNETR architecture. It combines a ViT with U-Net, replaces the convolutional encoder 
    with the ViT and adapt each skip connection signal to their layer's spatial dimensionality. 

    Reference: `UNETR: Transformers for 3D Medical Image Segmentation 
    <https://openaccess.thecvf.com/content/WACV2022/html/Hatamizadeh_UNETR_Transformers_for_3D_Medical_Image_Segmentation_WACV_2022_paper.html>`_.

    Parameters
    ----------
    input_shape : 3D/4D tuple
        Dimensions of the input image. E.g. ``(y, x, channels)`` or ``(z, y, x, channels)``.
        
    patch_size : int
        Size of the patches that are extracted from the input image. As an example, to use ``16x16`` 
        patches, set ``patch_size = 16``.

    embed_dim : int
        Dimension of the embedding space.

    depth : int
        Number of transformer encoder layers.

    num_heads : int
        Number of heads in the multi-head attention layer.

    mlp_ratio : float, optional
        Ratio to multiply ``embed_dim`` to obtain the dense layers of the final classifier. 

    num_filters: int, optional
        Number of filters in the first UNETR's layer of the decoder. In each layer the previous number of filters is 
        doubled.

    norm_layer : Torch layer, optional
        Nomarlization layer to use in ViT backbone.

    n_classes : int, optional
        Number of classes to predict. Is the number of channels in the output tensor.

    decoder_activation : str, optional
        Activation function for the decoder.

    ViT_hidd_mult : int, optional
        Multiple of the transformer encoder layers from of which the skip connection signal is going to be extracted.
        E.g. if we have ``12`` transformer encoder layers, and we set ``ViT_hidd_mult = 3``, we are going to take
        ``[1*ViT_hidd_mult, 2*ViT_hidd_mult, 3*ViT_hidd_mult]`` -> ``[Z3, Z6, Z9]`` encoder's signals. 

    batch_norm : bool, optional
        Whether to use batch normalization or not.

    dropout : bool, optional
        Dropout rate for the decoder (can be a list of dropout rates for each layer).

    output_channels : str, optional
        Channels to operate with. Possible values: ``BC``, ``BCD``, ``BP``, ``BCDv2``,
        ``BDv2``, ``Dv2`` and ``BCM``.

    Returns
    -------
    model : Torch model
        UNETR model.
    """
    def __init__(self, input_shape, patch_size, embed_dim, depth, num_heads, mlp_ratio=4., num_filters = 16, 
        norm_layer=nn.LayerNorm, n_classes = 1, decoder_activation = 'relu', ViT_hidd_mult = 3, batch_norm = True, 
        dropout = 0.0, output_channels="BC"):
        super().__init__()
        
        self.input_shape = input_shape
        self.embed_dim = embed_dim
        self.patch_size = patch_size
        self.ViT_hidd_mult = ViT_hidd_mult
        self.ndim = 3 if len(input_shape) == 4 else 2
        self.n_classes = 1 if n_classes <= 2 else n_classes
        self.multiclass = True if n_classes > 2 and output_channels is not None else False

        if self.ndim == 3:
            conv = nn.Conv3d
            convtranspose = nn.ConvTranspose3d
            batchnorm_layer = nn.BatchNorm3d if batch_norm else None
            self.reshape_shape = (
                self.input_shape[0]//self.patch_size, 
                self.input_shape[1]//self.patch_size,
                self.input_shape[2]//self.patch_size,
                self.embed_dim,
            )
            self.permutation = (0, 4, 1, 2, 3)
        else:
            conv = nn.Conv2d
            convtranspose = nn.ConvTranspose2d
            batchnorm_layer = nn.BatchNorm2d if batch_norm else None
            self.reshape_shape = (
                self.input_shape[0]//self.patch_size, 
                self.input_shape[1]//self.patch_size,
                self.embed_dim,
            )
            self.permutation = (0, 3, 1, 2)

        # ViT part
        self.patch_embed = PatchEmbed(img_size=input_shape[0], patch_size=patch_size, in_chans=input_shape[-1],
            ndim=self.ndim, embed_dim=embed_dim)
        num_patches = self.patch_embed.num_patches

        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + 1, embed_dim), requires_grad=False)  # fixed sin-cos embedding

        self.blocks = nn.ModuleList([
            Block(embed_dim, num_heads, mlp_ratio, qkv_bias=True, norm_layer=norm_layer)
            for i in range(depth)])
        self.norm = norm_layer(embed_dim)

        # UNETR Part (bottom_up, from the bottle-neck, to the output)
        self.total_upscale_factor = int(math.log2(patch_size))
        # make a list of dropout values if needed
        if type( dropout ) is float: 
            dropout = [dropout,]*self.total_upscale_factor
        
        self.bottleneck = convtranspose(embed_dim, num_filters * (2**(self.total_upscale_factor-1)), kernel_size=2, stride=2, bias=False)

        self.mid_blue_block = nn.ModuleList()
        self.two_yellow_layers = nn.ModuleList()
        self.up_green_layers = nn.ModuleList()
        for layer in reversed(range(1, self.total_upscale_factor)):
            block = [] 
            in_size = embed_dim
            for _ in range(self.total_upscale_factor - layer):
                block.append(
                    convtranspose(in_size, num_filters * (2**layer), kernel_size=2, stride=2, bias=False)
                )
                block.append(
                    ConvBlock(conv, in_size=num_filters * (2**layer), out_size=num_filters * (2**layer), k_size=3, 
                        act=decoder_activation, batch_norm=batchnorm_layer, dropout=dropout[layer])
                )
                in_size = num_filters * (2**layer)
            self.mid_blue_block.append(nn.Sequential(*block))
            self.two_yellow_layers.append(
                DoubleConvBlock(conv, in_size*2, in_size, k_size=3, act=decoder_activation, batch_norm=batchnorm_layer,
                    dropout=dropout[layer]))
            self.up_green_layers.append(
                convtranspose(in_size, num_filters * (2**(layer-1)), kernel_size=2, stride=2, bias=False))
        
        # Last two yellow block for the first skip connection 
        self.two_yellow_layers.append(
            DoubleConvBlock(conv, input_shape[-1], num_filters, k_size=3, act=decoder_activation, batch_norm=batchnorm_layer,
                dropout=dropout[0]))

        # Last convolutions 
        self.two_yellow_layers.append(
            DoubleConvBlock(conv, num_filters*2, num_filters, k_size=3, act=decoder_activation, batch_norm=batchnorm_layer,
                dropout=dropout[0]))

        # Instance segmentation
        if output_channels is not None:
            if output_channels == "Dv2":
                self.last_block = conv(num_filters, 1, kernel_size=1, padding='same')
            elif output_channels in ["BC", "BP"]:
                self.last_block = conv(num_filters, 2, kernel_size=1, padding='same')
            elif output_channels in ["BDv2", "BD"]:
                self.last_block = conv(num_filters, 2, kernel_size=1, padding='same')
            elif output_channels in ["BCM", "BCD", "BCDv2"]:
                self.last_block = conv(num_filters, 3, kernel_size=1, padding='same')
        # Other
        else:
            self.last_block = conv(num_filters, self.n_classes, kernel_size=1, padding='same')

        # Multi-head: instances + classification
        self.last_class_head = None
        if self.multiclass:
            self.last_class_head = conv(num_filters, self.n_classes, kernel_size=1, padding='same')

        self.apply(self._init_weights)
        
    def proj_feat(self, x):
        x = x.view((x.size(0),) + self.reshape_shape)
        x = x.permute(self.permutation).contiguous()
        return x

    def forward(self, input)-> torch.Tensor | List[torch.Tensor]:
        # Vit part
        B = input.shape[0]  
        x = self.patch_embed(input)

        cls_tokens = self.cls_token.expand(B, -1, -1) 
        x = torch.cat((cls_tokens, x), dim=1)
        x = x + self.pos_embed

        skip_connection_index = [self.ViT_hidd_mult * layer for layer in range(1, self.total_upscale_factor)]
        skip_connections = []
        for i, blk in enumerate(self.blocks):
            x = blk(x)    
            if (i+1) in skip_connection_index:
                skip_connections.insert(0, x[:, 1:, :])

        # CNN Decoder 
        x = self.bottleneck(self.proj_feat(x[:, 1:, :]))

        for i, layers in enumerate(zip(self.mid_blue_block,self.two_yellow_layers,self.up_green_layers)):
            blue, yellow, green = layers
            z = self.proj_feat(skip_connections[i])
            z = blue(z)
            x = torch.cat([x, z], dim=1)
            x = yellow(x)
            x = green(x)
            
        # first skip connection (out of transformer)
        first_skip = self.two_yellow_layers[-2](input)
        x = torch.cat([first_skip, x], dim=1)
        
        # UNETR output 
        x = self.two_yellow_layers[-1](x)
        class_head_out = torch.empty(())    
        if self.multiclass and self.last_class_head is not None:
            class_head_out = self.last_class_head(x) 
        x = self.last_block(x)
        
        if self.multiclass:
            return [x, class_head_out]
        else:
            return x

    def _init_weights(self, m):
        if isinstance(m, nn.Conv2d) or isinstance(m, nn.Conv3d):
            nn.init.xavier_uniform_(m.weight)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.Linear):
            nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)




import torch
import math
import sys
import numpy as np
from typing import Iterable
from timm.utils import accuracy

from biapy.utils.misc import MetricLogger, SmoothedValue, all_reduce_mean, to_pytorch_format

def train_one_epoch(cfg, model, model_call_func, loss_function, activations, metric_function, prepare_targets, data_loader, optimizer, 
    device, loss_scaler, epoch, log_writer=None, lr_scheduler=None, start_steps=0, verbose=False):

    model.train(True)

    # Ensure correct order of each epoch info by adding loss first
    metric_logger = MetricLogger(delimiter="  ", verbose=verbose)
    metric_logger.add_meter('loss', SmoothedValue())

    header = 'Epoch: [{}]'.format(epoch+1)
    print_freq = 10

    optimizer.zero_grad()
                        
    for step, (batch, targets) in enumerate(metric_logger.log_every(data_loader, print_freq, header)):

        # Apply warmup cosine decay scheduler if selected
        # (notice we use a per iteration (instead of per epoch) lr scheduler)
        if epoch % cfg.TRAIN.ACCUM_ITER == 0 and cfg.TRAIN.LR_SCHEDULER.NAME == 'warmupcosine':
            lr_scheduler.adjust_learning_rate(optimizer, step / len(data_loader) + epoch)

        it = start_steps + step  # global training iteration

        # Gather inputs
        targets = prepare_targets(targets, batch)

        # Pass the images through the model
        # TODO: control autocast and mixed precision
        with torch.cuda.amp.autocast(enabled=False):
            outputs = activations(model_call_func(batch, is_train=True), training=True)
            loss = loss_function(outputs, targets)

        loss_value = loss.item()
        if not math.isfinite(loss_value):
            print("Loss is {}, stopping training".format(loss_value))
            sys.exit(1)

        # Calculate the metrics
        metric_function(outputs, targets, metric_logger)

        # Forward pass scaling the loss
        loss /= cfg.TRAIN.ACCUM_ITER
        if (step + 1) % cfg.TRAIN.ACCUM_ITER == 0:
            loss.backward()
            optimizer.step() #update weight        
            optimizer.zero_grad()
            if lr_scheduler is not None and cfg.TRAIN.LR_SCHEDULER.NAME == 'onecycle':
                lr_scheduler.step() 

        if device.type != 'cpu':
            torch.cuda.synchronize()

        # Update loss in loggers
        metric_logger.update(loss=loss_value)
        loss_value_reduce = all_reduce_mean(loss_value)
        if log_writer is not None: log_writer.update(loss=loss_value_reduce, head="loss")

        # Update lr in loggers
        max_lr = 0.
        for group in optimizer.param_groups:
            max_lr = max(max_lr, group["lr"])
        if step == 0:
            metric_logger.add_meter('lr', SmoothedValue(window_size=1, fmt='{value:.6f}'))
        metric_logger.update(lr=max_lr)
        if log_writer is not None: log_writer.update(lr=max_lr, head="opt")
        
        # Update weight_decay in loggers
        if verbose:
            weight_decay_value = None
            for group in optimizer.param_groups:
                if group["weight_decay"] > 0:
                    weight_decay_value = group["weight_decay"]
            metric_logger.update(weight_decay=weight_decay_value)
            log_writer.update(weight_decay=weight_decay_value, head="opt")

    # Gather the stats from all processes
    metric_logger.synchronize_between_processes()
    print("[Train] averaged stats:", metric_logger)
    return {k: meter.global_avg for k, meter in metric_logger.meters.items()}


@torch.no_grad()
def evaluate(cfg, model, model_call_func, loss_function, activations, metric_function, prepare_targets, epoch, 
    data_loader, lr_scheduler):
    metric_logger = MetricLogger(delimiter="  ")
    header = 'Epoch: [{}]'.format(epoch+1)

    # Switch to evaluation mode
    model.eval()

    for batch in metric_logger.log_every(data_loader, 10, header):
        # Gather inputs
        images = batch[0]
        targets = batch[1]
        targets = prepare_targets(targets, images)

        # Pass the images through the model
        # TODO: control autocast and mixed precision
        with torch.cuda.amp.autocast(enabled=False):  
            outputs = activations(model_call_func(images, is_train=True), training=True)
            loss = loss_function(outputs, targets)
        
        # Calculate the metrics
        metric_function(outputs, targets, metric_logger)
    
        metric_logger.update(loss=loss.item())

    # Gather the stats from all processes
    metric_logger.synchronize_between_processes()

    print("[Val] averaged stats:", metric_logger)

    # Apply reduceonplateau scheduler if the global validation has been reduced
    if lr_scheduler is not None and cfg.TRAIN.LR_SCHEDULER.NAME == 'reduceonplateau':
        lr_scheduler.step(metric_logger.meters['loss'].global_avg)
    return {k: meter.global_avg for k, meter in metric_logger.meters.items()}



import random
import torch
from torch.utils.data import Dataset
import os
import h5py
import numpy as np
from skimage.io import imread
from PIL import Image
from PIL.TiffTags import TAGS

from biapy.data.pre_processing import normalize, norm_range01


class test_pair_data_generator(Dataset):
    """
    Image data generator without data augmentation. Used only for test data.

    Parameters
    ----------
    X : Numpy 5D/4D array, optional
        Data. E.g. ``(num_of_images, z, y, x, channels)`` for ``3D`` or ``(num_of_images, y, x, channels)`` for ``2D``.

    d_path : Str, optional
        Path to load the data from.

    test_by_chunks : bool, optional
        Tell the generator that the data is going to be read by chunks and by H5/Zarr files. 

    provide_Y: bool, optional
        Whether to return ground truth, using ``Y`` or loading from ``dm_path``.

    Y : Numpy 5D/4D array, optional
        Data mask. E.g. ``(num_of_images, z, y, x, channels)`` for ``3D`` or ``(num_of_images, y, x, channels)`` for ``2D``.

    dm_path : Str, optional
        Path to load the mask data from.

    dims: str, optional
        Dimension of the data. Possible options: ``2D`` or ``3D``.

    seed : int, optional
        Seed for random functions.

    instance_problem : bool, optional
        To not divide the labels if being in an instance segmenation problem.

    norm_type : str, optional
        Type of normalization to be made. Options available: ``div`` or ``custom``.

    not_normalize : bool, optional
        Whether to normalize the data or not. Useful in BMZ model as the normalization is made during the inference. 

    norm_custom_mean : float, optional
        Mean of the data used to normalize.

    norm_custom_std : float, optional
        Std of the data used to normalize.

    reduce_mem : bool, optional
        To reduce the dtype from float32 to float16. 

    sample_ids :  List of ints, optional
        When cross validation is used specific training samples are passed to the generator.
    
    convert_to_rgb : bool, optional
        In case RGB images are expected, e.g. if ``crop_shape`` channel is 3, those images that are grayscale are 
        converted into RGB.
    """
    def __init__(self, ndim, X=None, d_path=None, test_by_chunks=False, provide_Y=False, Y=None, dm_path=None, seed=42,
                 instance_problem=False, normalizeY='as_mask', norm_type='div', not_normalize=False,
                 norm_custom_mean=None, norm_custom_std=None, reduce_mem=False, sample_ids=None, convert_to_rgb=False):

        if X is None and d_path is None:
            raise ValueError("One between 'X' or 'd_path' must be provided")
        if provide_Y:
            if Y is None and dm_path is None:
                raise ValueError("One between 'Y' or 'dm_path' must be provided")
        assert normalizeY in ['as_mask', 'as_image', 'none']
        
        self.X = X
        self.Y = Y
        self.d_path = d_path
        self.dm_path = dm_path
        self.test_by_chunks = test_by_chunks
        self.provide_Y = provide_Y
        self.convert_to_rgb = convert_to_rgb

        if not reduce_mem:
            self.dtype = np.float32  
            self.dtype_str = "float32"
        else:
            self.dtype = np.float16
            self.dtype_str = "float16"
        self.data_path = sorted(next(os.walk(d_path))[2]) if X is None else None
        if sample_ids is not None and self.data_path is not None:
            self.data_path = [x for i, x in enumerate(self.data_path) if i in sample_ids]
        if provide_Y:
            self.data_mask_path = sorted(next(os.walk(dm_path))[2]) if Y is None else None
            if sample_ids is not None and self.data_mask_path is not None:
                self.data_mask_path = [x for i, x in enumerate(self.data_mask_path) if i in sample_ids]
                
            if self.data_path is not None and self.data_mask_path is not None:
                if len(self.data_path) != len(self.data_mask_path):
                    raise ValueError("Different number of raw and ground truth items ({} vs {}). "
                        "Please check the data!".format(len(self.data_path), len(self.data_mask_path)))
        self.seed = seed
        self.ndim = ndim
        if X is None:
            self.len = len(self.data_path)
            if len(self.data_path) == 0:
                if test_by_chunks:
                    print("No image found in {} folder. Assumming that files are zarr directories.")
                    self.data_path = sorted(next(os.walk(d_path))[1])
                    if provide_Y:
                        self.data_mask_path = sorted(next(os.walk(dm_path))[1])
                    if len(self.data_path) == 0:
                        raise ValueError("No zarr files found in {}".format(d_path))
                else:
                    raise ValueError("No test image found in {}".format(d_path))
        else:
            self.len = len(X)
        self.o_indexes = np.arange(self.len)
        self.normalizeY = normalizeY
        
        self.not_normalize = not_normalize
        # Check if a division is required
        self.X_norm = {}
        self.X_norm['type'] = 'div'
        if provide_Y:
            self.Y_norm = {}
            self.Y_norm['type'] = 'div'
        img, mask, xnorm = self.load_sample(0)

        if norm_type == 'custom' and not not_normalize:
            if norm_custom_mean is not None and norm_custom_std is not None:
                self.X_norm['mean'] = norm_custom_mean
                self.X_norm['std'] = norm_custom_std
            self.X_norm['orig_dtype'] = img.dtype
            self.X_norm['type'] = 'custom'
        if xnorm:
            self.X_norm.update(xnorm)

        if mask is not None:
            self.Y_norm = {}
            if normalizeY == 'as_mask':
                self.Y_norm['type'] = 'div'
                if (np.max(mask) > 30 and not instance_problem):
                    self.Y_norm['div'] = 1   
            elif normalizeY == 'as_image':
                self.Y_norm.update(self.X_norm)

    def norm_X(self, img):
        """
        X data normalization.

        Parameters
        ----------
        img : 3D/4D Numpy array
            X element, for instance, an image. E.g. ``(y, x, channels)`` in ``2D`` and ``(z, y, x, channels)`` in ``3D``.

        Returns
        -------
        img : 3D/4D Numpy array
            X element normalized. E.g. ``(y, x, channels)`` in ``2D`` and ``(z, y, x, channels)`` in ``3D``.
        
        xnorm : dict, optional
            Normalization info. 
        """
        xnorm = None
        if not self.not_normalize:
            if self.X_norm['type'] == 'div':
                img, xnorm = norm_range01(img, dtype=self.dtype)
            elif self.X_norm['type'] == 'custom':
                img = normalize(img, self.X_norm['mean'], self.X_norm['std'], out_type=self.dtype_str)
        return img, xnorm

    def norm_Y(self, mask):   
        """
        Y data normalization.

        Parameters
        ----------
        mask : 3D/4D Numpy array
            Y element, for instance, an image's mask. E.g. ``(y, x, channels)`` in ``2D`` and ``(z, y, x, channels)`` in 
            ``3D``.

        Returns
        -------
        mask : 3D/4D Numpy array
            Y element normalized. E.g. ``(y, x, channels)`` in ``2D`` and ``(z, y, x, channels)`` in ``3D``.
        
        ynorm : dict, optional
            Normalization info.
        """
        ynorm = None
        if self.normalizeY == 'as_mask':
            if 'div' in self.Y_norm:
                mask = mask/255
        elif self.normalizeY == 'as_image':
            if self.X_norm['type'] == 'div':
                mask, xnorm = norm_range01(mask, dtype=self.dtype)
            elif self.X_norm['type'] == 'custom':
                mask = normalize(mask, self.X_norm['mean'], self.X_norm['std'], out_type=self.dtype_str)
        return mask, ynorm

    def load_sample(self, idx):
        """Load one data sample given its corresponding index."""
        mask = None
        # Choose the data source
        if self.X is None:
            if self.data_path[idx].endswith('.npy'):
                img = np.load(os.path.join(self.d_path, self.data_path[idx]))
                if self.provide_Y:
                    mask = np.load(os.path.join(self.dm_path, self.data_mask_path[idx]))
            elif self.data_path[idx].endswith('.hdf5') or self.data_path[idx].endswith('.h5'):
                if not self.test_by_chunks:
                    img = h5py.File(os.path.join(self.d_path, self.data_path[idx]),'r')
                    img = img[list(img)[0]]
                    if self.provide_Y:
                        mask = h5py.File(os.path.join(self.dm_path, self.data_mask_path[idx]),'r')
                        mask = mask[list(mask)[0]]
                else:
                    img = os.path.join(self.d_path, self.data_path[idx])
                    if self.provide_Y:
                        mask = os.path.join(self.dm_path, self.data_mask_path[idx])
            elif self.data_path[idx].endswith('.zarr'):
                if self.test_by_chunks:
                    img = os.path.join(self.d_path, self.data_path[idx])
                    if self.provide_Y:
                        mask = os.path.join(self.dm_path, self.data_mask_path[idx])
                else:
                   raise NotImplementedError
            else:
                img = imread(os.path.join(self.d_path, self.data_path[idx]))
                img = np.squeeze(img)
                if self.provide_Y:
                    mask = imread(os.path.join(self.dm_path, self.data_mask_path[idx]))
                    mask = np.squeeze(mask)  
        else:
            img = self.X[idx]
            img = np.squeeze(img)

            if self.provide_Y:
                mask = self.Y[idx]
                mask = np.squeeze(mask)

        # Correct dimensions 
        if not self.test_by_chunks:
            if self.ndim == 3:
                if img.ndim == 3: 
                    img = np.expand_dims(img, -1)
                else:
                    min_val = min(img.shape)
                    channel_pos = img.shape.index(min_val)
                    if channel_pos != 3 and img.shape[channel_pos] <= 4:
                        new_pos = [x for x in range(4) if x != channel_pos]+[channel_pos,]
                        img = img.transpose(new_pos)
            else:
                if img.ndim == 2: 
                    img = np.expand_dims(img, -1) 
                else:
                    if img.shape[0] <= 3: img = img.transpose((1,2,0))
            if self.provide_Y:
                if self.ndim == 3:
                    if mask.ndim == 3: 
                        mask = np.expand_dims(mask, -1)
                    else:
                        min_val = min(mask.shape)
                        channel_pos = mask.shape.index(min_val)
                        if channel_pos != 3 and mask.shape[channel_pos] <= 4:
                            new_pos = [x for x in range(4) if x != channel_pos]+[channel_pos,]
                            mask = mask.transpose(new_pos)
                else:
                    if mask.ndim == 2: 
                        mask = np.expand_dims(mask, -1)
                    else:
                        if mask.shape[0] <= 3: mask = mask.transpose((1,2,0))

        xnorm = self.X_norm 
        if not self.test_by_chunks:
            # Normalization
            img, xnorm = self.norm_X(img)
            if self.provide_Y:
                mask, xnorm = self.norm_Y(mask)

            img = np.expand_dims(img, 0).astype(self.dtype)
            if self.provide_Y:
                mask = np.expand_dims(mask, 0)
                if self.normalizeY == 'as_mask':
                    mask = mask.astype(np.uint8)

        if self.convert_to_rgb and img.shape[-1] == 1:
            img = np.repeat(img, 3, axis=-1)
            
        return img, mask, xnorm


    def __len__(self):
        """Defines the length of the generator"""
        return self.len


    def __getitem__(self, index):
        """Generation of one pair of data.

           Parameters
           ----------
           index : int
               Sample index counter.

           Returns
           -------
           img : 3D/4D Numpy array
               X element, for instance, an image. E.g. ``(z, y, x, channels)`` if ``2D`` or 
               ``(y, x, channels)`` if ``3D``. 
               
           mask : 3D/4D Numpy array
               Y element, for instance, a mask. E.g. ``(z, y, x, channels)`` if ``2D`` or 
               ``(y, x, channels)`` if ``3D``.
        """
        img, mask, norm = self.load_sample(index)
        
        if norm is not None:
            self.X_norm.update(norm)
                    
        if self.provide_Y:
            return img, self.X_norm, mask, self.Y_norm
        else:
            return img, self.X_norm

    def get_data_normalization(self):
        return self.X_norm



import os
import torch
import scipy
import numpy as np
from tqdm import tqdm
import pandas as pd
from skimage.segmentation import clear_border, find_boundaries
from skimage.io import imread
from scipy.ndimage.morphology import binary_dilation  as binary_dilation_scipy
from scipy.ndimage.measurements import center_of_mass                                                  
from skimage.morphology import disk, dilation, erosion, binary_erosion, binary_dilation
from skimage.measure import label, regionprops
from skimage.transform import resize
from skimage.feature import peak_local_max, canny
from skimage.exposure import equalize_adapthist
from skimage.color import rgb2gray
from skimage.filters import gaussian, median

from biapy.utils.util import load_data_from_dir, load_3d_images_from_dir, save_tif


#########################
# INSTANCE SEGMENTATION #
#########################
def create_instance_channels(cfg, data_type='train'):
    """Create training and validation new data with appropiate channels based on ``PROBLEM.INSTANCE_SEG.DATA_CHANNELS`` for instance
       segmentation.

       Parameters
       ----------
       cfg : YACS CN object
           Configuration.

	   data_type: str, optional
		   Wheter to create training or validation instance channels.

       Returns
       -------
       filenames: List of str
           Image paths.
    """

    assert data_type in ['train', 'val']

    f_name = load_data_from_dir if cfg.PROBLEM.NDIM == '2D' else load_3d_images_from_dir
    tag = "TRAIN" if data_type == "train" else "VAL"
    Y, _, _, filenames = f_name(getattr(cfg.DATA, tag).GT_PATH, check_drange=False, return_filenames=True)
    print("Creating Y_{} channels . . .".format(data_type))
    if isinstance(Y, list):
        for i in tqdm(range(len(Y))):
            if cfg.MODEL.N_CLASSES > 2:
                if Y[i].shape[-1] != 2:
                    raise ValueError("In instance segmentation, when 'MODEL.N_CLASSES' are more than 2 labels need to have two channels, "
                        "e.g. (256,256,2), containing the instance segmentation map (first channel) and classification map (second channel).")
                else:
                    class_channel = np.expand_dims(Y[i,...,1].copy(),-1)
                    Y[i] = labels_into_channels(Y[i], mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS, save_dir=getattr(cfg.PATHS, tag+'_INSTANCE_CHANNELS_CHECK'),
                        fb_mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CONTOUR_MODE)
                    Y[i] = np.concatenate([Y[i],class_channel],axis=-1)
            else:
                Y[i] = labels_into_channels(Y[i], mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS, save_dir=getattr(cfg.PATHS, tag+'_INSTANCE_CHANNELS_CHECK'),
                    fb_mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CONTOUR_MODE)
    else:
        if cfg.MODEL.N_CLASSES > 2:
            if Y.shape[-1] != 2:
                raise ValueError("In instance segmentation, when 'MODEL.N_CLASSES' are more than 2 labels need to have two channels, "
                    "e.g. (256,256,2), containing the instance segmentation map (first channel) and classification map (second channel).")
            else:
                class_channel = np.expand_dims(Y[...,1].copy(),-1)
                Y = labels_into_channels(Y, mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS, save_dir=getattr(cfg.PATHS, tag+'_INSTANCE_CHANNELS_CHECK'),
                    fb_mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CONTOUR_MODE)
                Y = np.concatenate([Y, class_channel], axis=-1)
        else:
            Y = labels_into_channels(Y, mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS, save_dir=getattr(cfg.PATHS, tag+'_INSTANCE_CHANNELS_CHECK'),
                fb_mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CONTOUR_MODE)
    
    save_tif(Y, data_dir=getattr(cfg.DATA, tag).INSTANCE_CHANNELS_MASK_DIR, filenames=filenames, verbose=cfg.TEST.VERBOSE)
    X, _, _, filenames = f_name(getattr(cfg.DATA, tag).PATH, return_filenames=True)
    print("Creating X_{} channels . . .".format(data_type))
    save_tif(X, data_dir=getattr(cfg.DATA, tag).INSTANCE_CHANNELS_DIR, filenames=filenames, verbose=cfg.TEST.VERBOSE)
    
    # Save original X data with the labels 
    for i in range(min(3,len(X))):
        if isinstance(X, list):
            save_tif(X[i], getattr(cfg.PATHS, tag+'_INSTANCE_CHANNELS_CHECK'), filenames=['vol'+str(i)+".tif"], verbose=False)
        else:
            save_tif(np.expand_dims(X[i],0), getattr(cfg.PATHS, tag+'_INSTANCE_CHANNELS_CHECK'), filenames=['vol'+str(i)+".tif"], verbose=False)

def create_test_instance_channels(cfg):
    """Create test new data with appropiate channels based on ``PROBLEM.INSTANCE_SEG.DATA_CHANNELS`` for instance segmentation.

       Parameters
       ----------
       cfg : YACS CN object
           Configuration.
    """

    f_name = load_data_from_dir if cfg.PROBLEM.NDIM == '2D' else load_3d_images_from_dir

    if cfg.DATA.TEST.LOAD_GT:
        Y_test, _, _, test_filenames = f_name(cfg.DATA.TEST.GT_PATH, check_drange=False, return_filenames=True)
        print("Creating Y_test channels . . .")
        if isinstance(Y_test, list):
            for i in tqdm(range(len(Y_test))):
                Y_test[i] = labels_into_channels(Y_test[i], mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS, save_dir=cfg.PATHS.TEST_INSTANCE_CHANNELS_CHECK,
                                            fb_mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CONTOUR_MODE)
        else:
            Y_test = labels_into_channels(Y_test, mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS, save_dir=cfg.PATHS.TEST_INSTANCE_CHANNELS_CHECK,
                                     fb_mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CONTOUR_MODE)
        save_tif(Y_test, data_dir=cfg.DATA.TEST.INSTANCE_CHANNELS_MASK_DIR, filenames=test_filenames, verbose=cfg.TEST.VERBOSE)

    print("Creating X_test channels . . .")
    X_test, _, _, test_filenames = f_name(cfg.DATA.TEST.PATH, return_filenames=True)
    save_tif(X_test, data_dir=cfg.DATA.TEST.INSTANCE_CHANNELS_DIR, filenames=test_filenames, verbose=cfg.TEST.VERBOSE)
    
    # Save original X data with the labels 
    for i in range(min(3,len(X_test))):
        if isinstance(X_test, list):
            save_tif(X_test[i], cfg.PATHS.TEST_INSTANCE_CHANNELS_CHECK, filenames=['vol'+str(i)+".tif"], verbose=True)
        else:
            save_tif(np.expand_dims(X_test[i],0), cfg.PATHS.TEST_INSTANCE_CHANNELS_CHECK, filenames=['vol'+str(i)+".tif"], verbose=True)

def labels_into_channels(data_mask, mode="BC", fb_mode="outer", save_dir=None):
    """Converts input semantic or instance segmentation data masks into different binary channels to train an instance segmentation
       problem. 

       Parameters
       ----------
       data_mask : 4D/5D Numpy array
           Data mask to create the new array from. It is expected to have just one channel. E.g. ``(10, 200, 1000, 1000, 1)``

       mode : str, optional
           Operation mode. Possible values: ``BC`` and ``BCD``.  ``BC`` corresponds to use binary segmentation+contour.
           ``BCD`` stands for binary segmentation+contour+distances.

       fb_mode : str, optional
          Mode of the find_boundaries function from ``scikit-image`` or "dense". More info in:
          `find_boundaries() <https://scikit-image.org/docs/stable/api/skimage.segmentation.html#skimage.segmentation.find_boundaries>`_. 
          Choose "dense" to label as contour every pixel that is not in ``B`` channel. 

       save_dir : str, optional
           Path to store samples of the created array just to debug it is correct.

       Returns
       -------
       new_mask : 5D Numpy array
           5D array with 3 channels instead of one. E.g. ``(10, 200, 1000, 1000, 3)``
    """

    assert data_mask.ndim in [5, 4]
    assert mode in ['BC', 'BCM', 'BCD', 'BD', 'BCDv2', 'Dv2', 'BDv2', 'BP']

    d_shape = 4 if data_mask.ndim == 5 else 3
    if mode in ['BCDv2', 'Dv2', 'BDv2']:
        c_number = 4
    elif mode in ['BCD', 'BCM']:
        c_number = 3
    elif mode in ['BC', 'BP', 'BD']:
        c_number = 2

    if 'D' in mode:
        dtype = np.float32  
    else:
        dtype = np.uint8

    new_mask = np.zeros(data_mask.shape[:d_shape] + (c_number,), dtype=dtype)
    for img in tqdm(range(data_mask.shape[0])):
        vol = data_mask[img,...,0].astype(np.int64)
        instances = np.unique(vol)
        instance_count = len(instances)

        # Background distance
        if 'Dv2' in mode:
            # Background distance
            vol_b_dist = np.invert(vol>0)
            vol_b_dist= scipy.ndimage.distance_transform_edt(vol_b_dist)
            vol_b_dist = np.max(vol_b_dist)-vol_b_dist
            new_mask[img,...,3] = vol_b_dist.copy()

        # Semantic mask
        if 'B' in mode and instance_count != 1:
            new_mask[img,...,0] = (vol>0).copy().astype(np.uint8)

        # Central points
        if 'P' in mode and instance_count != 1:
            coords = center_of_mass(vol>0, vol, instances[1:])
            coords = np.round(coords).astype(int)
            for coord in coords:
                if data_mask.ndim == 5:
                    z,y,x = coord
                    new_mask[img,z,y,x,1] = 1
                else:
                    y,x = coord
                    new_mask[img,y,x,1] = 1

            if data_mask.ndim == 5:
                for i in range(new_mask.shape[1]):                                                                                  
                    new_mask[img,i,...,1] = dilation(new_mask[img,i,...,1], disk(3)) 
            else:
                new_mask[img,...,1] = dilation(new_mask[img,...,1], disk(3))  

        # Contour
        if ('C' in mode or 'Dv2' in mode) and instance_count != 1: 
            f = "thick" if fb_mode == "dense" else fb_mode
            new_mask[img,...,1] = find_boundaries(vol, mode=f).astype(np.uint8)
            if fb_mode == "dense" and mode != "BCM":
                new_mask[img,...,1] = 1 - binary_dilation(new_mask[img,...,1], disk(1))
                new_mask[img,...,1] = 1 - ( (vol>0) * new_mask[img,...,1])
            if 'B' in mode:
                # Remove contours from segmentation maps
                new_mask[img,...,0][np.where(new_mask[img,...,1] == 1)] = 0
            if mode == "BCM":
                new_mask[img,...,2] = (vol>0).astype(np.uint8)

        if ('D' in mode or 'Dv2' in mode) and instance_count != 1:
            # Foreground distance
            new_mask[img,...,-1] = scipy.ndimage.distance_transform_edt(new_mask[img,...,0])
            props = regionprops(vol, new_mask[img,...,-1])
            max_values = np.zeros(vol.shape)
            for i in range(len(props)):
                max_values = np.where(vol==props[i].label, props[i].intensity_max, max_values)     
            new_mask[img,...,-1] = max_values - new_mask[img,...,-1]

    # Normalize and merge distance channels
    if 'Dv2' in mode:
        # Normalize background   
        b_min = np.min(new_mask[...,3])
        b_max = np.max(new_mask[...,3])    
        new_mask[...,3] = (new_mask[...,3]-b_min)/(b_max-b_min)

        if instance_count != 1: 
            # Normalize foreground  
            f_min = np.min(new_mask[...,2])
            f_max = np.max(new_mask[...,2])
            new_mask[...,2] = (new_mask[...,2]-f_min)/(f_max-f_min)

            new_mask[...,2] = new_mask[...,3] - new_mask[...,2]

            # The intersection of the channels is the contour channel, so set it to the maximum value 1
            new_mask[...,2][new_mask[...,1]>0] = 1

        if mode == "BCDv2":
            new_mask = new_mask[...,:3]
        elif mode == "BDv2":
            new_mask = new_mask[...,[0,-1]]   
        elif mode == "Dv2":
            new_mask = np.expand_dims(new_mask[...,2], -1)

    if save_dir is not None:
        os.makedirs(save_dir, exist_ok=True)
        suffix = []
        if mode == "Dv2":
            suffix.append('_distance.tif')
        else:
            suffix.append('_semantic.tif')
        if mode in ["BC", "BCM", "BCD", "BCDv2"]:
            suffix.append('_contour.tif')
            if mode in ["BCD", "BCDv2"]:
                suffix.append('_distance.tif')
            elif mode == "BCM":
                suffix.append('_binary_mask.tif')
        elif mode == "BP":
            suffix.append('_points.tif')
        elif mode in ["BDv2", "BD"]:
            suffix.append('_distance.tif')

        for i in range(min(3,len(new_mask))):
            for j in range(len(suffix)):
                aux = new_mask[i,...,j]
                aux = np.expand_dims(np.expand_dims(aux,-1),0)
                save_tif(aux, save_dir, filenames=['vol'+str(i)+suffix[j]], verbose=False)
                save_tif(np.expand_dims(data_mask[i],0), save_dir, filenames=['vol'+str(i)+"_y.tif"], verbose=False)
    return new_mask

#############
# DETECTION #
#############
def create_detection_masks(cfg, data_type='train'):
    """Create detection masks based on CSV files.

       Parameters
       ----------
       cfg : YACS CN object
           Configuration.

	   data_type: str, optional
		   Wheter to create train, validation or test masks.
    """

    assert data_type in ['train', 'val', 'test']

    if data_type == "train":
        tag = "TRAIN"
    elif data_type == "val":
        tag = "VAL"
    else:
        tag = "TEST"
    img_dir = getattr(cfg.DATA, tag).PATH
    label_dir = getattr(cfg.DATA, tag).GT_PATH
    out_dir = getattr(cfg.DATA, tag).DETECTION_MASK_DIR
    img_ids = sorted(next(os.walk(img_dir))[2])
    img_ext = '.'+img_ids[0].split('.')[-1]
    ids = sorted(next(os.walk(label_dir))[2])
    if len(img_ids) != len(ids):
        raise ValueError("Different number of CSV files and images found ({} vs {}). "
            "Please check that every image has one and only one CSV file".format(len(ids), len(img_ids)))
    if cfg.PROBLEM.NDIM == '2D':
        req_dim = 2 
        req_columns = ['axis-0', 'axis-1']
        req_columns_class = ['axis-0', 'axis-1', 'class']
    else:
        req_dim = 3
        req_columns = ['axis-0', 'axis-1', 'axis-2']
        req_columns_class = ['axis-0', 'axis-1', 'axis-2', 'class']

    classes = cfg.MODEL.N_CLASSES if cfg.MODEL.N_CLASSES > 2 else 1
    print("Creating {} detection masks . . .".format(data_type))
    for i in range(len(ids)):
        img_filename = os.path.splitext(ids[i])[0]+img_ext
        if not os.path.exists(os.path.join(out_dir, img_filename)):
            print("Attempting to create mask from CSV file: {}".format(os.path.join(label_dir, ids[i])))
            if not os.path.exists(os.path.join(img_dir, img_filename)):
                print("WARNING: The image seems to have different name than its CSV file. Using the CSV file that's "
                "in the same spot (within the CSV files list) where the image is in its own list of images. Check if it is correct!")
                img_filename = img_ids[i]
            print("Its respective image seems to be: {}".format(os.path.join(img_dir, img_filename)))
            
            df = pd.read_csv(os.path.join(label_dir, ids[i]))  
            img = imread(os.path.join(img_dir, img_filename))
            
            # Adjust shape
            img = np.squeeze(img)
            if cfg.PROBLEM.NDIM == '2D':
                if img.ndim == 2:
                    img = np.expand_dims(img, -1)
                else:
                    if img.shape[0] <= 3: img = img.transpose((1,2,0))   
            else: 
                if img.ndim == 3: 
                    img = np.expand_dims(img, -1)
                else:
                    if img.shape[0] <= 3: img = img.transpose((1,2,3,0))

            # Discard first index column to not have error if it is not sorted 
            p_number=df.iloc[: , 0].to_list()
            df = df.iloc[: , 1:]
            df = df.rename(columns=lambda x: x.strip()) # trim spaces in column names
            if len(df.columns) == req_dim+1:
                if not all(df.columns == req_columns_class):
                    raise ValueError("CSV columns need to be {}".format(req_columns_class))
            elif len(df.columns) == req_dim:
                if not all(df.columns == req_columns):
                    raise ValueError("CSV columns need to be {}".format(req_columns))
            else:
                raise ValueError("CSV file {} need to have {} or {} columns. Found {}"
                                .format(os.path.join(label_dir, ids[i]), req_dim, req_dim+1, len(df.columns)))

            # Convert them to int in case they are floats
            df['axis-0'] = df['axis-0'].astype('int')
            df['axis-1'] = df['axis-1'].astype('int')
            if cfg.PROBLEM.NDIM == '3D':
                df['axis-2'] = df['axis-2'].astype('int')
            
            df = df.sort_values(by=['axis-0']) 

            # Obtain the points 
            z_axis_point = df['axis-0']                                                                       
            y_axis_point = df['axis-1']    
            if cfg.PROBLEM.NDIM == '3D':                                                                   
                x_axis_point = df['axis-2']
            
            # Class column present
            if len(df.columns) == req_dim+1:
                df['class'] = df['class'].astype('int')
                class_point = np.array(df['class']) 

                uniq = np.sort(np.unique(class_point))        
                if uniq[0] != 1:
                    raise ValueError("Class number must start with 1")    
                if not all(uniq == np.array(range(1,classes+1))):
                    raise ValueError("Classes must be consecutive, e.g [1,2,3,4..]. Given {}".format(uniq))   
            else:
                if classes > 1:
                    raise ValueError("MODEL.N_CLASSES > 1 but no class specified in CSV files (4th column must have class info)")
                class_point = [1] * len(z_axis_point)

            # Create masks
            print("Creating all points . . .")
            mask = np.zeros((img.shape[:-1]+(classes,)), dtype=np.uint8)
            for j in tqdm(range(len(z_axis_point)), total=len(z_axis_point), leave=False):
                a0_coord = z_axis_point[j]
                a1_coord = y_axis_point[j]
                if cfg.PROBLEM.NDIM == '3D':
                    a2_coord = x_axis_point[j]
                c_point = class_point[j]-1

                if c_point+1 > mask.shape[-1]:
                    raise ValueError("Class {} detected while MODEL.N_CLASSES was set to {}. Please check it!"
                        .format(c_point+1, classes))

                # Paint the point
                if cfg.PROBLEM.NDIM == '3D':
                    if a0_coord < mask.shape[0] and a1_coord < mask.shape[1] and a2_coord < mask.shape[2]:
                        cpd = cfg.PROBLEM.DETECTION.CENTRAL_POINT_DILATION                                                                                                              
                        if 1 in mask[max(0,a0_coord-1):min(mask.shape[0],a0_coord+2),                                   
                                     max(0,a1_coord-1-cpd):min(mask.shape[1],a1_coord+2+cpd),                                   
                                     max(0,a2_coord-1-cpd):min(mask.shape[2],a2_coord+2+cpd), c_point]: 
                            print("WARNING: possible duplicated point in (3,9,9) neighborhood: coords {} , class {} "
                                  "(point number {} in CSV)".format((a0_coord,a1_coord,a2_coord), c_point, p_number[j]))                                                                                                                                            
                                                                                                                                
                        mask[a0_coord,a1_coord,a2_coord,c_point] = 1                                            
                        if a1_coord+1 < mask.shape[1]: mask[a0_coord,a1_coord+1,a2_coord,c_point] = 1       
                        if a1_coord-1 > 0: mask[a0_coord,a1_coord-1,a2_coord,c_point] = 1                   
                        if a2_coord+1 < mask.shape[2]: mask[a0_coord,a1_coord,a2_coord+1,c_point] = 1       
                        if a2_coord-1 > 0: mask[a0_coord,a1_coord,a2_coord-1,c_point] = 1     
                        if cfg.PROBLEM.DETECTION.CENTRAL_POINT_DILATION == 0:
                            if a1_coord+1 < mask.shape[1] and a2_coord+1 < mask.shape[2]: 
                                mask[a0_coord,a1_coord+1,a2_coord+1,c_point] = 1       
                            if a1_coord-1 > 0 and a2_coord-1 > 0: 
                                mask[a0_coord,a1_coord-1,a2_coord-1,c_point] = 1  
                            if a1_coord-1 > 0 and a2_coord+1 < mask.shape[2]: 
                                mask[a0_coord,a1_coord-1,a2_coord+1,c_point] = 1   
                            if a1_coord+1 < mask.shape[1] and a2_coord-1 > 0: 
                                mask[a0_coord,a1_coord+1,a2_coord-1,c_point] = 1             
                    else:  
                        print("WARNING: discarding point {} which seems to be out of shape: {}"
                              .format([a0_coord,a1_coord,a2_coord], img.shape))                                                                                                     
                else:
                    if a0_coord < mask.shape[0] and a1_coord < mask.shape[1]:                                                                                                              
                        if 1 in mask[max(0,a0_coord-4):min(mask.shape[0],a0_coord+5),                                   
                                    max(0,a1_coord-4):min(mask.shape[1],a1_coord+5), c_point]: 
                            print("WARNING: possible duplicated point in (9,9) neighborhood: coords {} , class {} "
                                  "(point number {} in CSV)".format((a0_coord,a1_coord), c_point, p_number[j]))                                                                                                                                            
                                                                                                                                
                        mask[a0_coord,a1_coord,c_point] = 1                                            
                        if a1_coord+1 < mask.shape[1]: mask[a0_coord,a1_coord+1,c_point] = 1       
                        if a1_coord-1 > 0: mask[a0_coord,a1_coord-1,c_point] = 1                                     
                    else:  
                        print("WARNING: discarding point {} which seems to be out of shape: {}"
                              .format([a0_coord,a1_coord], img.shape))     

            # Dilate the mask
            if cfg.PROBLEM.DETECTION.CENTRAL_POINT_DILATION > 0:
                if cfg.PROBLEM.NDIM == '2D': mask = np.expand_dims(mask,0)
                for k in range(mask.shape[0]): 
                    for ch in range(mask.shape[-1]):                                                                                  
                        mask[k,...,ch] = binary_dilation_scipy(mask[k,...,ch], iterations=1,  structure=disk(cfg.PROBLEM.DETECTION.CENTRAL_POINT_DILATION))                                                                                                                                                    
                if cfg.PROBLEM.NDIM == '2D': mask = mask[0]

            if cfg.PROBLEM.DETECTION.CHECK_POINTS_CREATED:
                print("Check points created to see if some of them are very close that create a large label") 
                error_found = False
                for ch in range(mask.shape[-1]):
                    _, index, counts = np.unique(label(clear_border(mask[...,ch])), return_counts=True, return_index=True)                     
                    # 0 is background so valid element is 1. We will compare that value with the rest                                                                         
                    ref_value = counts[1]                                                                                           
                    for k in range(2,len(counts)):                                                                                  
                        if abs(ref_value - counts[k]) > 5:                                                                          
                            point = np.unravel_index(index[k], mask[...,ch].shape)  
                            print("WARNING: There is a point (coords {}) with size very different from "
                                  "the rest. Maybe that cell has several labels: please check it! Normally all point "
                                  "have {} pixels but this one has {}.".format(point, ref_value, counts[k])) 
                            error_found = True

                if error_found:
                    raise ValueError("Duplicate points have been found so please check them before continuing. "
                                     "If you consider that the points are valid simply disable "
                                     "'PROBLEM.DETECTION.CHECK_POINTS_CREATED' so this check is not done again!")

            save_tif(np.expand_dims(mask,0), out_dir, [img_filename])
        else:
            print("Mask file {} found for CSV file: {}".format(os.path.join(out_dir, img_filename), 
                os.path.join(label_dir, ids[i])))

#######
# SSL #
#######
def create_ssl_source_data_masks(cfg, data_type='train'):
    """Create SSL source data.

       Parameters
       ----------
       cfg : YACS CN object
           Configuration.

	   data_type: str, optional
		   Wheter to create train, validation or test source data.
    """

    assert data_type in ['train', 'val', 'test']
    tag = data_type.upper()
    
    img_dir = getattr(cfg.DATA, tag).PATH
    out_dir = getattr(cfg.DATA, tag).SSL_SOURCE_DIR
    ids = sorted(next(os.walk(img_dir))[2])
    add_noise = True if cfg.PROBLEM.SELF_SUPERVISED.NOISE > 0 else False

    print("Creating {} SSL source. . .".format(data_type))
    for i in range(len(ids)):
        if not os.path.exists(os.path.join(out_dir, ids[i])):
            print("Crappifying file {} to create SSL source".format(os.path.join(img_dir, ids[i])))

            img = imread(os.path.join(img_dir, ids[i]))
            
            # Adjust shape
            img = np.squeeze(img)
            if cfg.PROBLEM.NDIM == '2D':
                if img.ndim == 2:
                    img = np.expand_dims(img, -1)
                else:
                    if img.shape[0] <= 3: img = img.transpose((1,2,0))   
            else: 
                if img.ndim == 3: 
                    img = np.expand_dims(img, -1)
                else:
                    min_val = min(img.shape)
                    channel_pos = img.shape.index(min_val)
                    if channel_pos != 3 and img.shape[channel_pos] <= 4:
                        new_pos = [x for x in range(4) if x != channel_pos]+[channel_pos,]
                        img = img.transpose(new_pos)

            img = crappify(img, resizing_factor=cfg.PROBLEM.SELF_SUPERVISED.RESIZING_FACTOR, 
                add_noise=add_noise, noise_level=cfg.PROBLEM.SELF_SUPERVISED.NOISE)

            save_tif(np.expand_dims(img,0), out_dir, [ids[i]])
        else:
            print("Source file {} found".format(os.path.join(img_dir, ids[i])))

def crappify(input_img, resizing_factor, add_noise=True, noise_level=None, Down_up=True):
    """
    Crappifies input image by adding Gaussian noise and downsampling and upsampling it so the resolution
    gets worsen. 

    input_img : 4D/5D Numpy array
        Data to be modified. E.g. ``(y, x, channels)`` if working with 2D images or
        ``(z, y, x, channels)`` if working with 3D.

    resizing_factor : floats
        Downsizing factor to reshape the image.

    add_noise : boolean, optional
        Indicating whether to add gaussian noise before applying the resizing.

    noise_level: float, optional
        Number between ``[0,1]`` indicating the std of the Gaussian noise N(0,std).

    Down_up : bool, optional
        Indicating whether to perform a final upsampling operation to obtain an image of the 
        same size as the original but with the corresponding loss of quality of downsizing and 
        upsizing.

    Returns
    -------
    img : 4D/5D Numpy array
        Train images. E.g. ``(y, x, channels)`` if working with 2D images or
        ``(z, y, x, channels)`` if working with 3D.
    """
    if input_img.ndim == 3:
        w, h, c = input_img.shape
        org_sz = (w, h)
    else:
        d, w, h, c = input_img.shape
        org_sz = (d, w, h)
        new_d = int(d / np.sqrt(resizing_factor))

    new_w = int(w / np.sqrt(resizing_factor))
    new_h = int(h / np.sqrt(resizing_factor))

    if input_img.ndim == 3:
        targ_sz = (new_w, new_h)
    else:
        targ_sz = (new_d, new_w, new_h)
    
    img = input_img.copy()
    if add_noise:
        img = add_gaussian_noise(img, noise_level)

    img = resize(img, targ_sz, order=1, mode='reflect',
                 clip=True, preserve_range=True, anti_aliasing=False)

    if Down_up:
        img = resize(img, org_sz, order=1, mode='reflect',
                        clip=True, preserve_range=True, anti_aliasing=False)

    return img.astype(input_img.dtype)

def add_gaussian_noise(image, percentage_of_noise):
    """
    Adds Gaussian noise to an input image. 

    Parameters
    ----------
    image : 3D Numpy array
        Image to be added Gaussian Noise with 0 mean and a certain std. E.g. ``(y, x, channels)``.

    percentage_of_noise : float
        percentage of the maximum value of the image that will be used as the std of the Gaussian Noise 
        distribution.

    Returns
    -------
    out : 3D Numpy array
        Transformed image. E.g. ``(y, x, channels)``.
    """
    max_value=np.max(image)
    noise_level=percentage_of_noise*max_value
    noise = np.random.normal(loc=0, scale=noise_level, size=image.shape)
    noisy_img=np.clip(image+noise, 0, max_value).astype(image.dtype)
    return noisy_img

################
# SEMANTIC SEG #
################
def calculate_2D_volume_prob_map(Y, Y_path=None, w_foreground=0.94, w_background=0.06, save_dir=None):
    """Calculate the probability map of the given 2D data.

       Parameters
       ----------
       Y : 4D Numpy array
           Data to calculate the probability map from. E. g. ``(num_of_images, y, x, channel)``

       Y_path : str, optional
           Path to load the data from in case ``Y=None``.

       w_foreground : float, optional
           Weight of the foreground. This value plus ``w_background`` must be equal ``1``.

       w_background : float, optional
           Weight of the background. This value plus ``w_foreground`` must be equal ``1``.

       save_dir : str, optional
           Path to the file where the probability map will be stored.

       Raises
       ------
       ValueError
           if ``Y`` does not have 4 dimensions.

       ValueError
           if ``w_foreground + w_background > 1``.

       Returns
       -------
       Array : Str or 4D Numpy array
           Path where the probability map/s is/are stored if ``Y_path`` was given and there are images of different
           shapes. Otherwise, an array that represents the probability map of ``Y`` or all loaded data files from
           ``Y_path`` will be returned.
    """

    if Y is not None:
        if Y.ndim != 4:
            raise ValueError("'Y' must be a 4D Numpy array")

    if Y is None and Y_path is None:
        raise ValueError("'Y' or 'Y_path' need to be provided")

    if Y is not None:
        prob_map = np.copy(Y).astype(np.float32)
        l = prob_map.shape[0]
        channels = prob_map.shape[-1]
        v = np.max(prob_map)
    else:
        prob_map, _, _ = load_data_from_dir(Y_path)
        l = len(prob_map)
        channels = prob_map[0].shape[-1]
        v = np.max(prob_map[0])

    if isinstance(prob_map, list):
        first_shape = prob_map[0][0].shape
    else:
        first_shape = prob_map[0].shape

    print("Connstructing the probability map . . .")
    maps = []
    diff_shape = False
    for i in tqdm(range(l)):
        if isinstance(prob_map, list):
            _map = prob_map[i][0].copy().astype(np.float32)
        else:
            _map = prob_map[i].copy().astype(np.float32)

        for k in range(channels):
            # Remove artifacts connected to image border
            _map[:,:,k] = clear_border(_map[:,:,k])

            foreground_pixels = (_map[:,:,k] == v).sum()
            background_pixels = (_map[:,:,k] == 0).sum()

            if foreground_pixels == 0:
                _map[:,:,k][np.where(_map[:,:,k] == v)] = 0
            else:
                _map[:,:,k][np.where(_map[:,:,k] == v)] = w_foreground/foreground_pixels
            if background_pixels == 0:
                _map[:,:,k][np.where(_map[:,:,k] == 0)] = 0
            else:
                _map[:,:,k][np.where(_map[:,:,k] == 0)] = w_background/background_pixels

            # Necessary to get all probs sum 1
            s = _map[:,:,k].sum()
            if s == 0:
                t = 1
                for x in _map[:,:,k].shape: t *=x
                _map[:,:,k].fill(1/t)
            else:
                _map[:,:,k] = _map[:,:,k]/_map[:,:,k].sum()

        if first_shape != _map.shape: diff_shape = True
        maps.append(_map)

    if save_dir is not None:
        os.makedirs(save_dir, exist_ok=True)

        if not diff_shape:
            for i in range(len(maps)):
                maps[i] = np.expand_dims(maps[i], 0)
            maps = np.concatenate(maps)
            print("Saving the probability map in {}".format(save_dir))
            np.save(os.path.join(save_dir, 'prob_map.npy'), maps)
            return maps
        else:
            print("As the files loaded have different shapes, the probability map for each one will be stored"
                  " separately in {}".format(save_dir))
            d = len(str(l))
            for i in range(l):
                f = os.path.join(save_dir, 'prob_map'+str(i).zfill(d)+'.npy')
                np.save(f, maps[i])
            return save_dir

def calculate_3D_volume_prob_map(Y, Y_path=None, w_foreground=0.94, w_background=0.06, save_dir=None):
    """Calculate the probability map of the given 3D data.

       Parameters
       ----------
       Y : 5D Numpy array
           Data to calculate the probability map from. E. g. ``(num_subvolumes, z, y, x, channel)``

       Y_path : str, optional
           Path to load the data from in case ``Y=None``.

       w_foreground : float, optional
           Weight of the foreground. This value plus ``w_background`` must be equal ``1``.

       w_background : float, optional
           Weight of the background. This value plus ``w_foreground`` must be equal ``1``.

       save_dir : str, optional
           Path to the directory where the probability map will be stored.

       Returns
       -------
       Array : Str or 5D Numpy array
           Path where the probability map/s is/are stored if ``Y_path`` was given and there are images of different
           shapes. Otherwise, an array that represents the probability map of ``Y`` or all loaded data files from
           ``Y_path`` will be returned.

       Raises
       ------
       ValueError
           if ``Y`` does not have 5 dimensions.
       ValueError
           if ``w_foreground + w_background > 1``.
    """

    if Y is not None:
        if Y.ndim != 5:
            raise ValueError("'Y' must be a 5D Numpy array")

    if Y is None and Y_path is None:
        raise ValueError("'Y' or 'Y_path' need to be provided")

    if Y is not None:
        prob_map = np.copy(Y).astype(np.float32)
        l = prob_map.shape[0]
        channels = prob_map.shape[-1]
        v = np.max(prob_map)
    else:
        prob_map, _, _ = load_3d_images_from_dir(Y_path)
        l = len(prob_map)
        channels = prob_map[0].shape[-1]
        v = np.max(prob_map[0])

    if isinstance(prob_map, list):
        first_shape = prob_map[0][0].shape
    else:
        first_shape = prob_map[0].shape

    print("Constructing the probability map . . .")
    maps = []
    diff_shape = False
    for i in range(l):
        if isinstance(prob_map, list):
            _map = prob_map[i][0].copy().astype(np.float64)
        else:
            _map = prob_map[i].copy().astype(np.float64)

        for k in range(channels):
            for j in range(_map.shape[0]):
                # Remove artifacts connected to image border
                _map[j,:,:,k] = clear_border(_map[j,:,:,k])
            foreground_pixels = (_map[:,:,:,k] == v).sum()
            background_pixels = (_map[:,:,:,k] == 0).sum()

            if foreground_pixels == 0:
                _map[:,:,:,k][np.where(_map[:,:,:,k] == v)] = 0
            else:
                _map[:,:,:,k][np.where(_map[:,:,:,k] == v)] = w_foreground/foreground_pixels
            if background_pixels == 0:
                _map[:,:,:,k][np.where(_map[:,:,:,k] == 0)] = 0
            else:
                _map[:,:,:,k][np.where(_map[:,:,:,k] == 0)] = w_background/background_pixels

            # Necessary to get all probs sum 1
            s = _map[:,:,:,k].sum()
            if s == 0:
                t = 1
                for x in _map[:,:,:,k].shape: t *=x
                _map[:,:,:,k].fill(1/t)
            else:
                _map[:,:,:,k] = _map[:,:,:,k]/_map[:,:,:,k].sum()

        if first_shape != _map.shape: diff_shape = True
        maps.append(_map)

    if save_dir is not None:
        os.makedirs(save_dir, exist_ok=True)
        if not diff_shape:
            for i in range(len(maps)):
                maps[i] = np.expand_dims(maps[i], 0)
            maps = np.concatenate(maps)
            print("Saving the probability map in {}".format(save_dir))
            np.save(os.path.join(save_dir, 'prob_map.npy'), maps)
            return maps
        else:
            print("As the files loaded have different shapes, the probability map for each one will be stored "
                  "separately in {}".format(save_dir))
            d = len(str(l))
            for i in range(l):
                f = os.path.join(save_dir, 'prob_map'+str(i).zfill(d)+'.npy')
                np.save(f, maps[i])
            return save_dir

###########
# GENERAL #
###########
def norm_range01(x, dtype=np.float32):
    norm_steps = {}
    norm_steps['orig_dtype'] = x.dtype

    if x.dtype == np.uint8 or x.dtype == torch.uint8:
        x = x/255
        norm_steps['div_255'] = 1
    else:
        if (isinstance(x, np.ndarray) and np.max(x) > 255) or \
            (torch.is_tensor(x) and torch.max(x) > 255):
            norm_steps['reduced_{}'.format(x.dtype)] = 1
            x = reduce_dtype(x, 0, 65535, out_min=0, out_max=1, out_type=dtype)
        elif (isinstance(x, np.ndarray) and np.max(x) > 2) or \
            (torch.is_tensor(x) and torch.max(x) > 2):
            x = x/255
            norm_steps['div_255'] = 1

    if torch.is_tensor(x):
        x = x.to(dtype)
    else:
        x = x.astype(dtype)
    return x, norm_steps

def undo_norm_range01(x, xnorm):
    if 'div_255' in xnorm:
        x = (x*255)
        if isinstance(x, np.ndarray): 
            x = x.astype(np.uint8)
        else:
            x = x.to(torch.uint8)
    reductions = [key for key, value in xnorm.items() if 'reduced' in key.lower()]
    if len(reductions)>0:
        reductions = reductions[0]
        reductions = reductions.replace('reduced_','')
        x = (x*65535)
        if isinstance(x, np.ndarray): 
            x = x.astype(eval("np.{}".format(reductions) ))
        else:
            x = x.to(eval("torch.{}".format(reductions) ))
    return x

def reduce_dtype(x, x_min, x_max, out_min=0, out_max=1, out_type=np.float32):
    if isinstance(x, np.ndarray):
        return ((np.array((x-x_min)/(x_max-x_min))*(out_max-out_min))+out_min).astype(out_type)
    else: # Tensor considered
        return ((((x-x_min)/(x_max-x_min))*(out_max-out_min))+out_min).to(out_type)

def normalize(data, means, stds, out_type="float32"):
    numpy_torch_dtype_dict = {
        "bool"       : [torch.bool, bool],
        "uint8"      : [torch.uint8, np.uint8],
        "int8"       : [torch.int8, np.int8],
        "int16"      : [torch.int16, np.int16],
        "int32"      : [torch.int32, np.int32],
        "int64"      : [torch.int64, np.int64],
        "float16"    : [torch.float16, np.float16],
        "float32"    : [torch.float32, np.float32],
        "float64"    : [torch.float64, np.float64],
        "complex64"  : [torch.complex64, np.complex64],
        "complex128" : [torch.complex128, np.complex128],
    }
    if torch.is_tensor(data):
        return ((data - means) / stds).to(numpy_torch_dtype_dict[out_type][0])
    else:
        return ((data - means) / stds).astype(numpy_torch_dtype_dict[out_type][1])

def denormalize(data, means, stds, out_type="float32"):
    numpy_torch_dtype_dict = {
        "bool"       : [torch.bool, bool],
        "uint8"      : [torch.uint8, np.uint8],
        "int8"       : [torch.int8, np.int8],
        "int16"      : [torch.int16, np.int16],
        "int32"      : [torch.int32, np.int32],
        "int64"      : [torch.int64, np.int64],
        "float16"    : [torch.float16, np.float16],
        "float32"    : [torch.float32, np.float32],
        "float64"    : [torch.float64, np.float64],
        "complex64"  : [torch.complex64, np.complex64],
        "complex128" : [torch.complex128, np.complex128],
    }
    if torch.is_tensor(data):
        return ((data * stds) + means).to(numpy_torch_dtype_dict[out_type][0])
    else:
        return ((data * stds) + means).astype(numpy_torch_dtype_dict[out_type][1])

def resize_images(images, **kwards):
    '''
    The function resizes all the images using the specified parameters or default values if not provided.
    
    Parameters
    ----------
    images: Numpy array or list of numpy arrays
        The `images` parameter is the list of all input images that you want to resize.
    
    output_shape: iterable
        Size of the generated output image. E.g. `(256,256)`

    (kwards): optional
        `skimage.transform.resize() <https://scikit-image.org/docs/stable/api/skimage.transform.html#skimage.transform.resize>`__
        parameters are also allowed. 
        
    Returns
    -------
    resized_images: Numpy array or list of numpy arrays
        The resized images. The returned data will use the same data type as the given `images`.
    
    '''

    resized_images = [resize(img, **kwards).astype(img.dtype) for img in images]
    if isinstance(images, np.ndarray):
        resized_images = np.array(resized_images, dtype=images.dtype)
    return resized_images

def apply_gaussian_blur(images, **kwards):
    '''
    The function applies a Gaussian blur to all images.
    
    Parameters
    ----------
    images: Numpy array or list of numpy arrays
        The input images on which the Gaussian blur will be applied.

    (kwards): optional
        `skimage.filters.gaussian() <https://scikit-image.org/docs/stable/api/skimage.filters.html#skimage.filters.gaussian>`__
        parameters are also allowed. 
    
    Returns
    -------
    blurred_images: Numpy array or list of numpy arrays
        A Gaussian blurred images. The returned data will use the same data type as the given `images`.
    
    '''
    def _process(image, **kwards):
        im = gaussian(image, **kwards) # returns 0-1 range
        if np.issubdtype(image.dtype, np.integer):
            im = im * np.iinfo(image.dtype).max 
        im = im.astype(image.dtype)
        return im

    blurred_images = [_process(img, **kwards) for img in images]
    if isinstance(images, np.ndarray):
        blurred_images = np.array(blurred_images, dtype=images.dtype)
    return blurred_images

def apply_median_blur(images, **kwards):
    '''
    The function applies a median blur filter to all images.
    
    Parameters
    ----------
    image: Numpy array or list of numpy arrays
        The input image on which the median blur operation will be applied.
    
    (kwards): optional
        `skimage.filters.median() <https://scikit-image.org/docs/stable/api/skimage.filters.html#skimage.filters.median>`__
        parameters are also allowed.

    Returns
    -------
    blurred_images: Numpy array or list of numpy arrays
        The median-blurred images. The returned data will use the same data type as the given `images`.
    
    '''

    blurred_images = [median(img, **kwards).astype(img.dtype) for img in images]
    if isinstance(images, np.ndarray):
        blurred_images = np.array(blurred_images, dtype=images.dtype)
    return blurred_images

def detect_edges(images, **kwards):
    '''
    The function `detect_edges` takes the 2D images as input, converts it to grayscale if necessary, and
    applies the Canny edge detection algorithm to detect edges in the image.
    
    Parameters
    ----------
    images: Numpy array or list of numpy arrays
        The list of all input images on which the edge detection will be performed. It can be either a color image with
        shape (height, width, 3) or a grayscale image with shape (height, width, 1).
    
    (kwards): optional
        `skimage.feature.canny() <https://scikit-image.org/docs/stable/api/skimage.feature.html#skimage.feature.canny>`__
        parameters are also allowed.

    Returns
    -------
    edges: Numpy array or list of numpy arrays
        The edges of the input images. The returned numpy arrays will be uint8, where background is black (0) and edges white (255).
        The returned data will use the same structure as the given `images` (list[numpy array] or numpy array).
    
    '''

    def to_gray(image):
        c = image.shape[-1]
        if c == 3:
            image = rgb2gray(image)
        elif c == 1:
            image = image[...,0]
        else:
            raise ValueError(f'Detect edges function does not allow given ammount of channels ({c} channels). '
                             'Only accepts grayscale and RGB 2D images (1 or 3 channels).')
        return image
    
    def set_uint8(image):
        im = image.astype(np.uint8)
        im = im[..., np.newaxis] # add channel dim
        im = im * 255
        return im

    edges = [set_uint8(canny(to_gray(img), **kwards)) for img in images]
    if isinstance(images, np.ndarray):
        edges = np.array(edges, dtype=np.uint8) 
    return edges

def _histogram_matching(source_imgs, target_imgs):
    '''
    Given a set of target images, it will obtain their mean histogram 
    and applies histogram matching to all images from sorce images.
    
    Parameters
    ----------
    source_imgs: Numpy array or list of numpy arrays
        The images of the source domain, to which the histogram matching is to be applied.

    target_imgs: Numpy array
        The target domain images, from which mean histogram will be obtained.
    
    Returns
    -------
    matched_images : list of numpy arrays
        A set of source images with target's histogram
    '''

    hist_mean,_ = np.histogram(target_imgs.ravel(), bins=np.arange(np.iinfo(target_imgs.dtype).max + 1))
    hist_mean = hist_mean / target_imgs.shape[0] # number of images
    
    # calculate normalized quantiles
    tmpl_size = np.sum(hist_mean) 
    tmpl_quantiles = np.cumsum(hist_mean) / tmpl_size

    del target_imgs
    
    # based on scikit implementation. 
    # source: https://github.com/scikit-image/scikit-image/blob/v0.18.0/skimage/exposure/histogram_matching.py#L22-L70
    def _match_cumulative_cdf(source, tmpl_quantiles):
        src_values, src_unique_indices, src_counts = np.unique(source.ravel(),
                                                            return_inverse=True,
                                                            return_counts=True)

        # calculate normalized quantiles
        src_size = source.size # number of pixels
        src_quantiles = np.cumsum(src_counts) / src_size # normalize
        interp_a_values = np.interp(src_quantiles, tmpl_quantiles, np.arange(len(tmpl_quantiles)))

        return interp_a_values[src_unique_indices].reshape(source.shape)

    #apply histogram matching
    results = [_match_cumulative_cdf(image, tmpl_quantiles).astype(image.dtype) for image in source_imgs ]
    return results

def apply_histogram_matching(images, reference_path, is_2d):
    '''
    The function returns the images with their histogram matched to the histogram of the reference images, 
    loaded from the given reference_path.
    
    Parameters
    ----------
    images: Numpy array or list of numpy arrays
        The list of input images whose histogram needs to be matched to the reference histogram. It should be a
        numpy array representing the image.

    reference_path: str
        The reference_path is the directory path to the reference images. From reference images, we will extract
        the reference histogram with which we want to match the histogram of the images. It represents 
        the desired distribution of pixel intensities in the output image.
    
    is_2d: bool, optional
        The value indicate if the data given in ``reference_path`` is 2D (``is_2d = True``) or 3D (``is_2d = False``).
        Defaults to True.
    
    Returns
    -------
    matched_images : Numpy array or list of numpy arrays
        The result of matching the histogram of the input images to the histogram of the reference image.
        The returned data will use the same data type as the given `images`.
    '''
    f_name = load_data_from_dir if is_2d else load_3d_images_from_dir
    references, *_ = f_name(reference_path)
    
    matched_images = _histogram_matching(images, references)
    if isinstance(images, np.ndarray):
        matched_images = np.array(matched_images, dtype=images.dtype)
    return matched_images

def apply_clahe(images, **kwards):
    '''
    The function applies Contrast Limited Adaptive Histogram Equalization (CLAHE) to an image and
    returns the result.
    
    Parameters
    ----------
    images: Numpy array or list of numpy arrays
        The list of input images that you want to apply the CLAHE (Contrast Limited Adaptive Histogram Equalization)
        algorithm to.

    (kwards): optional
        `skimage.exposure.equalize_adapthist() <https://scikit-image.org/docs/stable/api/skimage.exposure.html#skimage.exposure.equalize_adapthist>`__
        parameters are also allowed.
    
    Returns
    -------
    processed_images: Numpy array or list of numpy arrays
        The images after applying the Contrast Limited Adaptive Histogram Equalization (CLAHE) algorithm.
        The returned data will use the same data type as the given `images`.
    '''

    def _process(image, **kwards):
        im = equalize_adapthist(image, **kwards) # returns 0-1 range
        if np.issubdtype(image.dtype, np.integer):
            im = im * np.iinfo(image.dtype).max 
        im = im.astype(image.dtype)
        return im
    
    processed_images = [_process(img, **kwards) for img in images]
    if isinstance(images, np.ndarray):
        processed_images = np.array(processed_images, dtype=images.dtype)
    return processed_images

def preprocess_data(cfg, x_data=[], y_data=[], is_2d=True, is_y_mask=False):
    """
    The function preprocesses data by applying various image processing techniques.
    
    Parameters
    ----------
    cfg: dict
        The `cfg` parameter is a configuration object that contains various settings for
        preprocessing the data. It is used to control the behavior of different preprocessing techniques
        such as image resizing, blurring, histogram matching, etc.

    x_data: 4D/5D numpy array or list of 3D/4D numpy arrays, optional
        The input data (images) to be preprocessed. The first dimension must be the number of images. 
        E.g. ``(num_of_images, y, x, channels)`` or ``(num_of_images, z, y, x, channels)``.
        In case of using a list, the format of the images remains the same. Each item in the list
        corresponds to a different image.
    
    y_data: 4D/5D numpy array or list of 3D/4D numpy arrays, optional
        The target data that corresponds to the x_data. The first dimension must be the number of images. 
        E.g. ``(num_of_images, y, x, channels)`` or ``(num_of_images, z, y, x, channels)``.
        In case of using a list, the format of the images remains the same. Each item in the list
        corresponds to a different image.
      
    is_2d: bool, optional
        A boolean flag indicating whether the reference data for histogram matching is 2D or not. Defaults to True.
      
    is_y_mask: bool, optional
        is_y_mask is a boolean parameter that indicates whether the y_data is a mask or not. If
        it is set to True, the resize operation for y_data will use the nearest neighbor interpolation
        method (order=0), otherwise it will use the interpolation method specified in the cfg.RESIZE.ORDER
        parameter. Defaults to False.
    
    Returns
    -------
    x_data: 4D/5D numpy array or list of 3D/4D numpy arrays, optional
        Preprocessed data. The same structure and dimensionality of the given data will be returned.

    y_data: 4D/5D numpy array or list of 3D/4D numpy arrays, optional
        Preprocessed data. The same structure and dimensionality of the given data will be returned.
    """
    
    if cfg.RESIZE.ENABLE:
        print("Preprocessing: applying resize . . .")
        if len(x_data) > 0:
            x_data = resize_images(x_data,
                output_shape = cfg.RESIZE.OUTPUT_SHAPE,
                order = cfg.RESIZE.ORDER,
                mode = cfg.RESIZE.MODE,
                cval = cfg.RESIZE.CVAL,
                clip = cfg.RESIZE.CLIP,
                preserve_range = cfg.RESIZE.PRESERVE_RANGE,
                anti_aliasing = cfg.RESIZE.ANTI_ALIASING
            )
        if len(y_data) > 0:
            # if y is a mask, then use nearest
            y_order = 0 if is_y_mask else cfg.RESIZE.ORDER
            y_data = resize_images(y_data,
                output_shape = cfg.RESIZE.OUTPUT_SHAPE,
                order = y_order,
                mode = cfg.RESIZE.MODE,
                cval = cfg.RESIZE.CVAL,
                clip = cfg.RESIZE.CLIP,
                preserve_range = cfg.RESIZE.PRESERVE_RANGE,
                anti_aliasing = cfg.RESIZE.ANTI_ALIASING
            )

    if len(x_data) > 0:
        if cfg.GAUSSIAN_BLUR.ENABLE:
            print("Preprocessing: applying gaussian blur . . .")
            x_data = apply_gaussian_blur(x_data,
                sigma = cfg.GAUSSIAN_BLUR.SIGMA,
                mode = cfg.GAUSSIAN_BLUR.MODE,
                channel_axis = cfg.GAUSSIAN_BLUR.CHANNEL_AXIS,
            )
        if cfg.MEDIAN_BLUR.ENABLE:
            print("Preprocessing: applying median blur . . .")
            x_data = apply_median_blur(x_data,
                footprint = cfg.MEDIAN_BLUR.FOOTPRINT,
            )
        if cfg.MATCH_HISTOGRAM.ENABLE:
            print("Preprocessing: applying histogram matching . . .")
            x_data = apply_histogram_matching(x_data,
                reference_path = cfg.MATCH_HISTOGRAM.REFERENCE_PATH,
                is_2d = is_2d,
            )
        if cfg.CLAHE.ENABLE:
            print("Preprocessing: applying CLAHE . . .")
            x_data = apply_clahe(x_data,
                kernel_size = cfg.CLAHE.KERNEL_SIZE,
                clip_limit = cfg.CLAHE.CLIP_LIMIT,
            )
        if cfg.CANNY.ENABLE:
            print("Preprocessing: applying Canny . . .")
            x_data = detect_edges(x_data,
                low_threshold = cfg.CANNY.LOW_THRESHOLD,
                high_threshold = cfg.CANNY.HIGH_THRESHOLD,
            )
    
    if len(x_data) > 0 and len(y_data) > 0:
        return x_data, y_data
    if len(y_data) > 0:
        return y_data
    else:
        return x_data